# -*- python -*-
# $RCSfile: meshparamwidgets.py,v $
# $Revision: 1.67 $
# $Author: langer $
# $Date: 2007/01/09 22:32:21 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov.

# ParameterWidgets for things that depend on a Mesh, such as Fields,
# Fluxes, and Equations defined on the Mesh.

from oof2.SWIG.common import ooferror
from oof2.SWIG.common import switchboard
from oof2.SWIG.engine import equation
from oof2.SWIG.engine import field
from oof2.SWIG.engine import flux
from oof2.SWIG.engine import planarity
from oof2.common import debug
from oof2.common.IO.GUI import chooser
from oof2.common.IO.GUI import parameterwidgets
from oof2.common.IO.GUI import whowidget
from oof2.engine import component
from oof2.engine import mesh
from oof2.engine import skeletoncontext
from oof2.engine import subproblemcontext
from oof2.engine.IO import meshparameters
import gtk
import string


class MeshParamWidgetBase(parameterwidgets.ParameterWidget):
    # Base class for a widget that displays and allows choices from a
    # list of things from a Mesh (eg, defined Fields or active
    # Equations).  The meshfunc constructor argument is a function
    # that returns the list of things to display.  It's called with
    # the mesh as its first argument, so it can be a Mesh member
    # function.  self.chooser is the ChooserWidget that displays the
    # items returned by meshfunc.

    def __init__(self, param, whoclass, meshfunc, scope, name=None,
                 separator_func=None):
        self.meshfunc = meshfunc
        self.whoclass = whoclass
        self.chooser = chooser.ChooserWidget([], callback=self.chooserCB,
                                             name=name,
                                             separator_func=separator_func)
        parameterwidgets.ParameterWidget.__init__(self, self.chooser.gtk, scope)
        self.meshwidget = scope.findWidget(
            lambda w: isinstance(w, whowidget.WhoWidget)
            and w.whoclass is whoclass)
        if self.meshwidget is None:
            raise ooferror.ErrPyProgrammingError("Can't find WhoWidget for %s"
                                                 % `whoclass`)
        self.sbcallbacks = [
            switchboard.requestCallbackMain(self.meshwidget, self.update),
            switchboard.requestCallbackMain("mesh changed", self.meshChangeCB),
            switchboard.requestCallbackMain("subproblem changed",
                                            self.meshChangeCB)]

        self.update(interactive=0)
        self.set_value(param.value)
        
    def update(self, interactive):
        msh = self.getSource()
        self.vals = {None:None}
        if msh is not None:
            namelist = []
            for obj in self.meshfunc(msh):
                self.vals[obj.name()] = obj
                namelist.append(obj.name())
            self.chooser.update(namelist)
            self.widgetChanged(len(namelist) > 0, interactive)
        else:
            self.chooser.update([])
            self.widgetChanged(0, interactive)
    def chooserCB(self, *args):
        self.widgetChanged(1, interactive=1)
    def meshChangeCB(self, meshcontext):
        if self.meshwidget:             # we haven't been cleaned up
            if meshcontext is self.getSource() or meshcontext.getParent() is self.getSource():
                self.update(interactive=0)
    def getSource(self):
        meshname = self.meshwidget.get_value()
        try:
            return self.whoclass[meshname]
        except KeyError:
            return None
    def get_value(self):
        return self.vals[self.chooser.get_value()]
    def set_value(self, obj):
        if obj is not None and obj.name() in self.chooser.choices():
            self.chooser.set_state(obj.name())
            self.widgetChanged(1, interactive=0)
        
    def cleanUp(self):
        debug.mainthreadTest()
        self.meshwidget = None
        map(switchboard.removeCallback, self.sbcallbacks)
        parameterwidgets.ParameterWidget.cleanUp(self)

class MeshParamWidget(MeshParamWidgetBase):
    def __init__(self, param, meshfunc, scope, name=None, separator_func=None):
        MeshParamWidgetBase.__init__(self, param,
                                     mesh.meshes,
                                     meshfunc, scope, name, separator_func)

class SubProblemParamWidget(MeshParamWidgetBase):
    def __init__(self, param, meshfunc, scope, name=None, separator_func=None):
        MeshParamWidgetBase.__init__(self, param,
                                     subproblemcontext.subproblems,
                                     meshfunc, scope, name, separator_func)
    

# Widgets for quantities to which Invariants can be calculated should
# be subclasses of InvariandWidget.

class InvariandWidget: pass

# Widgets for quantities for which components can be extracted should
# be subclasses of IndexableWidget.

class IndexableWidget: pass

#############################

def meshfieldlister(meshctxt):          # meshfunc for FieldParameterWidgets
    # list fields and out-of-plane fields, if defined.
    try:
        compoundfields = meshctxt.all_subproblem_fields()
        flds = []
        for fld in compoundfields:
            flds.append(fld)
            if not meshctxt.getObject().in_plane(fld): 
                flds.append(fld.out_of_plane())
        return flds
    except:
        # This might be called with an unresolvable proxy if there
        # are no meshes.
        return []
        
class FieldParameterWidget(MeshParamWidget, InvariandWidget, IndexableWidget):
    def __init__(self, param, scope, name=None):
        if param.outofplane:
            flist = meshfieldlister
        else:
            flist = mesh.Mesh.all_subproblem_fields
        MeshParamWidget.__init__(self, param, flist, scope, name)
##        self.sbcallbacks.append(
##            switchboard.requestCallbackMain("field defined", self.fieldDefCB))
    def fieldDefCB(self, *args):
        self.update(interactive=False)

def _makeFieldWidget(param, scope):
    return FieldParameterWidget(param, scope, name=param.name)

meshparameters.FieldParameter.makeWidget = _makeFieldWidget

def subpfieldlister(subp):     # meshfunc for SubProblemFieldParameterWidgets
    # list fields and out-of-plane fields, if defined.
    try:
        compoundfields = subp.all_fields()
        flds = []
        for fld in compoundfields:
            flds.append(fld)
            if not subp.getParent().getObject().in_plane(fld):
                flds.append(fld.out_of_plane())
        return flds
    except:
        # This might be called with an unresolvable proxy if there
        # are no meshes.
        return []
        
class SubProblemFieldParameterWidget(SubProblemParamWidget, InvariandWidget,
                                     IndexableWidget):
    def __init__(self, param, scope, name=None):
        if param.outofplane:
            flist = subpfieldlister
        else:
            flist = subproblemcontext.SubProblemContext.all_fields
        SubProblemParamWidget.__init__(self, param, flist, scope, name)

def _makeSubPFieldWidget(param, scope):
    return FieldParameterWidget(param, scope, name=param.name)

meshparameters.SubProblemFieldParameter.makeWidget = _makeSubPFieldWidget

#############################

class FluxParameterWidget(MeshParamWidget, InvariandWidget, IndexableWidget):
    def __init__(self, param, scope, name=None):
        MeshParamWidget.__init__(self, param,
                                 mesh.Mesh.all_subproblem_fluxes,
                                 scope,
                                 name=name)
            
def _makeFluxWidget(param, scope):
    return FluxParameterWidget(param, scope, name=param.name)

meshparameters.FluxParameter.makeWidget = _makeFluxWidget

class SubProblemFluxParameterWidget(SubProblemParamWidget, InvariandWidget,
                                    IndexableWidget):
    def __init__(self, param, scope, name=None):
        SubProblemParamWidget.__init__(
            self, param,
            subproblemcontext.SubProblemContext.all_fluxes,
            scope,
            name=name)
            
def _makeSubPFluxWidget(param, scope):
    return SubProblemParamWidget(param, scope, name=param.name)

meshparameters.SubProblemFluxParameter.makeWidget = _makeSubPFluxWidget

#############################

class EquationParameterWidget(MeshParamWidget, IndexableWidget):
    def __init__(self, param, scope, name=None):
        eqfunc = mesh.Mesh.all_subproblem_equations
        MeshParamWidget.__init__(self, param, eqfunc, scope, name=name)

def _makeEquationWidget(param, scope):
    return EquationParameterWidget(param, scope, name=param.name)

meshparameters.EquationParameter.makeWidget = _makeEquationWidget

class SubProblemEquationParameterWidget(SubProblemParamWidget, IndexableWidget):
    def __init__(self, param, scope, name=None):
        eqfunc = subproblemcontext.SubProblemContext.all_equations
        SubProblemParamWidget.__init__(self, param, eqfunc, scope, name=name)

def _makeSubPEquationWidget(param, scope):
    return SubProblemEquationParameterWidget(param, scope, name=param.name)

meshparameters.SubProblemEquationParameter.makeWidget = _makeSubPEquationWidget

class EquationBCParameterWidget(MeshParamWidget, IndexableWidget):
    def __init__(self, param, scope, name=None):
        eqfunc = mesh.Mesh.all_subproblem_equations_bc
        MeshParamWidget.__init__(self, param, eqfunc, scope, name)

def _makeEquationBCWidget(param, scope):
    return EquationBCParameterWidget(param, scope, name=param.name)

meshparameters.EquationBCParameter.makeWidget = _makeEquationBCWidget

class SubProblemEquationBCParameterWidget(SubProblemParamWidget,
                                          IndexableWidget):
    def __init__(self, param, scope, name=None):
        eqfunc = subproblemcontext.SubProblemContext.all_equations_bc
        SubProblemParamWidget.__init__(self, param, eqfunc, scope, name)

def _makeSubPEquationBCWidget(param, scope):
    return SubProblemEquationBCParameterWidget(param, scope, name=param.name)

meshparameters.SubProblemEquationBCParameter.makeWidget = \
                                                     _makeSubPEquationBCWidget

#############################################
#############################################

class FieldIndexParameterWidget(parameterwidgets.ParameterWidget):
    def __init__(self, param, scope, name=None):
        debug.mainthreadTest()
        self.chooser = chooser.ChooserWidget([], callback=self.chooserCB,
                                             name=name)
        box = gtk.VBox()
        box.pack_start(self.chooser.gtk, expand=0, fill=0)
        parameterwidgets.ParameterWidget.__init__(self, box, scope)
        self.fieldwidget = scope.findWidget(
            lambda w: isinstance(w, IndexableWidget))
        self.sbcallback = switchboard.requestCallbackMain(self.fieldwidget,
                                                          self.fieldCB)
        self.notapplicable = gtk.Label('(Not Applicable)')
        self.notapplicable.set_alignment(0.0, 0.5)
        box.pack_start(self.notapplicable, expand=0, fill=0)
        self.nIndices = 0
        self.update()
##        if param.value is not None and param.value != '':
        if param.value in self.chooser.choices():
            self.chooser.set_state(param.value)
        self.widgetChanged(1, interactive=0)
    def chooserCB(self, *args):
        self.widgetChanged(1, interactive=1)
    def fieldCB(self, interactive):
        self.update()
        self.widgetChanged(1, interactive)
    def update(self):                   # field has changed
        itlist = []
        self.nIndices = 0
        field = self.fieldwidget.get_value()
        if field is not None:
            iterator = field.iterator_all()
            while not iterator.end():
                self.nIndices += 1
                it = iterator.cloneIndex()
                itrepr = it.shortrepr()
                itlist.append(itrepr)
                iterator.next()
        self.chooser.update(itlist)
        self.show()
    def show(self):
        debug.mainthreadTest()
        self.gtk.show()
        if self.nIndices > 1:
            self.chooser.show()
            self.notapplicable.hide()
        else:
            self.chooser.gtk.hide()
            self.notapplicable.show()
    def get_value(self):
        val = self.chooser.get_value()
        if val is None:
            return ''
        return val
    def cleanUp(self):
        parameterwidgets.ParameterWidget.cleanUp(self)
        switchboard.removeCallback(self.sbcallback)

def _makeFieldIndexParameterWidget(param, scope):
    return FieldIndexParameterWidget(param, scope, name=param.name)

meshparameters.FieldIndexParameter.makeWidget = _makeFieldIndexParameterWidget


############################################################

# Not a widget for listing things in a SubProblem, but for listing the
# SubProblems in a Mesh.  This is only a bit of a hack, since it's
# different than the other widgets in the class.  The value of a
# SubProblemParameter is the SubProblem's path (similar to a
# WhoWidget), whereas the value of the other types of MeshParamWidget
# is a real object.

class SubProblemWidget(MeshParamWidget):
    def __init__(self, param, scope, name=None):
        MeshParamWidget.__init__(self, param, mesh.Mesh.subproblems,
                                 scope, name)
        self.sbcallbacks.append(
            switchboard.requestCallbackMain(("new who", "SubProblem"),
                                            self.newSubProblemCB))
    def newSubProblemCB(self, subppath):
        self.update(interactive=False)
    def get_value(self):
        subp = MeshParamWidget.get_value(self)
        if subp:
            return subp.path()
    def set_value(self, path):
        try:
            MeshParamWidget.set_value(self, subproblemcontext.subproblems[path])
        except KeyError:
            pass

def _makeSubProblemWidget(param, scope):
    return SubProblemWidget(param, scope, name=param.name)

subproblemcontext.SubProblemParameter.makeWidget = _makeSubProblemWidget

############################################################

# Special widget for mesh boundary parameters.  As it turns out, these
# params take strings, not boundary objects, so we override a
# significant fraction of the foregoing machinery, which is really
# designed for parameters which take the actual object.

class MeshBoundaryParamWidget(MeshParamWidget):
    def __init__(self, param, scope, name=None):
        MeshParamWidget.__init__(self, param, _getSortedBdyNames,
                                 scope, name=name, separator_func=_bdysepfunc)
        self.sbcallbacks.append(
            switchboard.requestCallbackMain('mesh boundaries changed',
                                            self.newBdys) )
    def newBdys(self, msh):
        if msh is self.getSource():
            self.update(interactive=0)
    def update(self, interactive):
        msh = self.getSource()
        if msh is not None:
            namelist = self.meshfunc(msh)
            self.chooser.update(namelist)
            if namelist: # If list has nonzero length, widget is valid.
                self.widgetChanged(1, interactive)
        else:
            self.chooser.update([])
            self.widgetChanged(0, interactive)
    def set_value(self, value):
        if value in self.chooser.choices():
            self.chooser.set_state(value)
            self.widgetChanged(1, interactive=0)
    def get_value(self):
        return self.chooser.get_value()

def _makeBoundaryWidget(param, scope):
    return MeshBoundaryParamWidget(param, scope, name=param.name)

meshparameters.MeshBoundaryParameter.makeWidget = _makeBoundaryWidget

# The MeshBoundaryParamWidget puts _separator_proxy in the list of
# boundaries to divide the edge boundaries from the point boundaries.
# The ChooserWidget replace it with a real separator, using
# _bdysepfunc as the predicate.

_separator_proxy = "----------"

def _getSortedBdyNames(msh):
    return msh.edgeBoundaryNames() + [_separator_proxy] + \
           msh.pointBoundaryNames()

def _bdysepfunc(model, iter):
    return model[iter][0] == _separator_proxy





# Special cases for nontrivial edge and point boundaries.


        
class MeshEdgeBdyParamWidget(MeshBoundaryParamWidget):
    def __init__(self, param, scope, name=None):
        MeshParamWidget.__init__(self, param,
                                  mesh.Mesh.edgeBoundaryNames,
                                  scope, name=name)
        self.sbcallbacks.append(
            switchboard.requestCallbackMain('mesh boundaries changed',
                                            self.newBdys) )

def _makeEdgeBdyWidget(param, scope):
    return MeshEdgeBdyParamWidget(param, scope, name=param.name)

meshparameters.MeshEdgeBdyParameter.makeWidget = _makeEdgeBdyWidget




class MeshPointBdyParamWidget(MeshBoundaryParamWidget):
    def __init__(self, param, scope, name=None):
        MeshParamWidget.__init__(self, param,
                                  mesh.Mesh.pointBoundaryNames,
                                  scope, name=name)
        self.sbcallbacks.append(
            switchboard.requestCallbackMain('mesh boundaries changed',
                                        self.newBdys) )


def _makePointBdyWidget(param, scope):
    return MeshPointBdyParamWidget(param, scope, name=param.name)

meshparameters.MeshPointBdyParameter.makeWidget = _makePointBdyWidget
