# -*- python -*-
# $RCSfile: meshinfodisplay.py,v $
# $Revision: 1.20 $
# $Author: reida $
# $Date: 2006/05/10 21:08:46 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov.

from oof2.common import color
from oof2.common import primitives
from oof2.common import registeredclass
from oof2.common.IO import display
from oof2.common.IO import parameter
from oof2.common.IO import xmlmenudump

class MeshInfoDisplay(display.DisplayMethod):
    def __init__(self, query_color, peek_color, node_size,
                 element_width):
        self.query_color = query_color
        self.peek_color = peek_color
        self.colors = {"query": self.query_color, "peek": self.peek_color}
        self.node_size = node_size
        self.element_width = element_width
        display.DisplayMethod.__init__(self)
        self.drawFuncs = {"Element": self.drawElement,
                          "Node": self.drawNode}


    def draw(self, gfxwindow, device):
        toolbox = gfxwindow.getToolboxByName("Mesh_Info")
        # Drawing "queried" item.
        if toolbox.querier and toolbox.querier.object:
            self.drawFuncs[toolbox.querier.targetname](device, toolbox, 
                                                       toolbox.querier.object,
                                                       which="query")
        # Drawing "peeked" item.
        if toolbox.peeker and toolbox.peeker.objects.values():
            for objtype in toolbox.peeker.objects.keys():
                if toolbox.peeker.objects[objtype]:
                    self.drawFuncs[objtype](device, toolbox, 
                                            toolbox.peeker.objects[objtype],
                                            which="peek")

    def drawElement(self, device, toolbox, element, which="query"):
        device.set_lineColor(self.colors[which])
        device.set_lineWidth(self.element_width)
        node_iter = element.cornernode_iterator().exteriornode_iterator()
        p_list = [node.position() for node in node_iter]
        displaced_p_list = [
            toolbox.meshlayer.displaced_from_undisplaced(
            toolbox.gfxwindow, x) for x in p_list]
        for i in range(len(displaced_p_list)):
            p0 = displaced_p_list[i]
            p1 = displaced_p_list[(i+1)%len(displaced_p_list)]
            device.draw_segment(primitives.Segment(p0, p1))


    def drawNode(self, device, toolbox, node, which="query"):
        device.set_lineColor(self.colors[which])
        device.set_lineWidth(self.node_size)
        displaced_position = toolbox.meshlayer.displaced_from_undisplaced(
            toolbox.gfxwindow(), node.position())
        device.draw_dot(displaced_position)        

    def getTimeStamp(self, gfxwindow):
        toolbox = gfxwindow.getToolboxByName("Mesh_Info")
        if toolbox.querier and toolbox.peeker:
            return max(self.timestamp,
                       toolbox.querier.getTimeStamp(),
                       toolbox.peeker.getTimeStamp())
        elif toolbox.querier and not toolbox.peeker:
            return max(self.timestamp,
                       toolbox.querier.getTimeStamp())
        else:
            return self.timestamp
                
# This object should be created via the registration, and not
# directly via the initializer, because the registration creation
# method gives it a timestamp.

meshInfoDisplay = registeredclass.Registration(
    'Info',
    display.DisplayMethod,
    MeshInfoDisplay,
    params=[color.ColorParameter('query_color', color.RGBColor(0.0, 0.5, 1.0),
                                 tip="Color for the queried object."),
            color.ColorParameter('peek_color', color.RGBColor(1.0, 0.5, 0.5),
                                 tip="Color for the peeked object."),
            parameter.IntRangeParameter('node_size', (0,10), 3, tip="Node size."),
            parameter.IntRangeParameter('element_width', (0,10), 2,
                                        tip="Line thickness for element edge.")],
    ordering=4.0,
    layerordering=display.PointLike(100),
    whoclasses=('Mesh',),
    tip="Set display parameters for the decorations used by the Mesh Info toolbox.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/reg/meshinfodisplay.xml')
    )

display.predefineLayer('Mesh', '<topmost>', meshInfoDisplay())
