// -*- C++ -*-
// $RCSfile: csubproblem.C,v $
// $Revision: 1.8 $
// $Author: rlua $
// $Date: 2007/02/14 00:01:28 $

/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

#include <oofconfig.h>
#include <map>
#include <utility>		// for std::pair
#include <vector>

#include "common/IO/progressbar.h"
#include "common/printvec.h"	// debugging
#include "common/tostring.h"
#include "engine/cconjugate.h"
#include "engine/csrmatrix.h"
#include "engine/element.h"
#include "engine/elementnodeiterator.h"
#include "engine/femesh.h"
#include "engine/flux.h"
#include "engine/linearizedsystem.h"
#include "engine/meshiterator.h"
#include "engine/nodalequation.h"
#include "engine/node.h"
#include "engine/csubproblem.h"

typedef std::pair<FuncNode::FieldSet, FuncNode::EquationSet> FEPair;
typedef std::map<FEPair, std::vector<int> > LocalMapDict;


CSubProblem::CSubProblem()
  : rwlock(0),
    mesh(0),
    precomputeRequired(false),
    linearsystem(this),
    global_equation_map(0),
    n_active_eqn(0),
    n_active_field(0),
    local_time(0.0)
{}

CSubProblem::~CSubProblem() {}

CMicrostructure *CSubProblem::get_microstructure() const {
  return mesh->get_microstructure();
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// NodeIterator CSubProblem::node_iterator() const {
//   return NodeIterator(new SubProblemNodeIterator(this));
// }

// FuncNodeIterator CSubProblem::funcnode_iterator() const {
//   return FuncNodeIterator(new SubProblemFuncNodeIterator(this)); 
// }

// ElementIterator CSubProblem::element_iterator() const {
//   return ElementIterator(new SubProblemElementIterator(this));
// }

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

CSubProblem::EquationData::EquationData()
  : active(0),
    kin_active(0),
    dyn_active(0)
{}

void CSubProblem::activate_equation(Equation &eqn) {
  // Trace("CSubProblem::activate_equation " + eqn.name());
  int index = eqn.index();
  // Make sure that there's an EquationData entry for this Equation
  if(index >= int(eqndata.size())) {
    eqndata.resize(index+1);
  }
  if(!eqndata[index].active) {
    // Mark the Equation as active
    eqndata[index].active = true;
    // Offset of this Equation in the Nodes' NodalEquation lists
//     eqndata[index].active_index = n_active_eqn;
    // Total number of active NodalEquations at each Node.
    ++n_active_eqn;

    // Create NodalEquations for this Equation in each Node.  This
    // also updates the Nodes' EquationSets.
    for(FuncNodeIterator ni=funcnode_iterator(); !ni.end(); ++ni) {
      ni.node()->addEquation(mesh, eqn);
    }
    // Activate fluxes required for this Equation
    eqn.activate_fluxes(this);
  }
}

void CSubProblem::deactivate_equation(Equation &eqn) {
  if(is_active_equation(eqn)) {
    int index = eqn.index();
    eqndata[index].active = false;
    for(FuncNodeIterator ni=funcnode_iterator(); !ni.end(); ++ni) {
      ni.node()->removeEquation(mesh, eqn);
    }
    // perform garbage collection on the nodaleqn list
    mesh->clean_nodaleqn();
    --n_active_eqn;
     eqn.deactivate_fluxes(this);
  }
}

// TODO PLASTICITY: How is is_active_equation used?  Should it examine
// each Node's EquationSet, or just check the CSubProblem's
// EquationData.active?  This may need a closer look when we get
// automatically activated auxilliary equations, esp. in
// Flux::make_stiffness and related functions.  Most other invocations
// of is_active_equation are used to determine what equations the user
// has explicitly activated.

bool CSubProblem::is_active_equation(const Equation &eqn) const {
  std::vector<EquationData>::size_type index = eqn.index();
  return (index < eqndata.size()) && eqndata[index].active;
}

int CSubProblem::n_active_eqns() const {
  return n_active_eqn;
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// Kinetics stuff

void CSubProblem::kinetic_activate_equation(Equation &eqn) {
  if (eqn.get_kinetic() >0)
    eqndata[eqn.index()].kin_active = true;
  //    eqn.activate_kinetics();
}


void CSubProblem::kinetic_deactivate_equation(Equation &eqn) {
  if(is_kinetically_active_equation(eqn) && eqn.get_kinetic() >0)
    //eqn.deactivate_kinetics();
    eqndata[eqn.index()].kin_active = true;
}


bool CSubProblem::is_kinetically_active_equation(Equation &eqn) {
  return is_active_equation(eqn) && eqndata[eqn.index()].kin_active;
}

bool CSubProblem::kinetics_activated() {
  for(std::vector<Equation*>::size_type i=0; i< all_equations()->size(); i++) {
    Equation * eqn = (*all_equations())[i];
    if (is_kinetically_active_equation(*eqn))
      return true;
  }
  return false;
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// Dynamics stuff

void CSubProblem::activate_dynamics(Equation &eqn) {
  if (eqn.get_dynamics() >0)
    //eqn.activate_dynamics();
    eqndata[eqn.index()].dyn_active = true;
}


void CSubProblem::deactivate_dynamics(Equation &eqn) {
  if(is_dynamically_active_equation(eqn) && eqn.get_dynamics() >0)
    //eqn.deactivate_dynamics();
    eqndata[eqn.index()].dyn_active = false;
}


bool CSubProblem::is_dynamically_active_equation(Equation &eqn) {
  return is_active_equation(eqn) && eqndata[eqn.index()].dyn_active;
}

bool CSubProblem::dynamics_activated() {
  for(std::vector<Equation*>::size_type i=0; i<all_equations()->size(); i++) {
    Equation * eqn  = (*all_equations())[i];
    if (is_dynamically_active_equation(*eqn))
      return true;
  }
  return false;
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// Flux bookkeeping. 

void CSubProblem::activate_flux(const Flux &flux) {
  std::vector<int>::size_type index = flux.index();
  if(index >= active_flux.size()) {
    active_flux.resize(index+1, 0);
  }
  ++active_flux[index];
}

void CSubProblem::deactivate_flux(const Flux &flux) {
  std::vector<int>::size_type index = flux.index();
  if(index < active_flux.size()) {
    if(active_flux[index] > 0)
      --active_flux[index];
  }
}

bool CSubProblem::is_active_flux(const Flux &flux) const {
  std::vector<int>::size_type index = flux.index();
  if(index < active_flux.size()) {
    return active_flux[index] > 0;
  }
  return false;
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// Field stuff

// activate, etc, for Fields have to be routed through Field virtual
// functions, because they have to know if the Field is a
// CompoundField or not.  The routines that do the real work are
// FEMesh::do_activate, etc.

CSubProblem::FieldData::FieldData()
  : active(0),
    defined(0)
{}

Field &CSubProblem::define_field(Field &field) {
  // Trace("CSubProblem::define_field " + field.name());
  field.define(this);
  return field;
}

void CSubProblem::undefine_field(Field &field) {
//   Trace("CSubProblem::undefine_field " + field.name());
  field.undefine(this);
}

void CSubProblem::activate_field(Field &field) {
  // Trace("CSubProblem::activate_field " + field.name());
  field.activate(this);
}

void CSubProblem::deactivate_field(Field &field) {
//   Trace("CSubProblem::deactivate_field " + field.name());
  field.deactivate(this);
}

void CSubProblem::do_define_field(Field &field) {
  // Trace("CSubProblem::do_define_field " + field.name() + " " + to_string(field.index()));
  requirePrecompute();
  std::vector<FieldData>::size_type index = field.index();
  if(fielddata.size() <= index) {
    fielddata.resize(index+1);
  }
  FieldData &fdata = fielddata[index];
  if(!fdata.defined) {
    fdata.defined = true;
    for(FuncNodeIterator ni=funcnode_iterator(); !ni.end(); ++ni) {
      ni.node()->addField(mesh, field);
    }
  }
}

void CSubProblem::do_undefine_field(Field &field) {
//   Trace("CSubProblem::do_undefine_field " + field.name())
  requirePrecompute();
  if(is_defined_field(field)) {
    if(is_active_field(field)) {
      deactivate_field(field);
    }
    std::vector<FieldData>::size_type index = field.index();
    if(index < fielddata.size()) {
      fielddata[index].defined = false;
      for(FuncNodeIterator ni=funcnode_iterator(); !ni.end(); ++ni) {
	ni.node()->removeField(mesh, field);
      }
      mesh->clean_doflist();
    }
  }
}

// Function to copy a field from another CSubProblem.  The other
// CSubProblem must have the same geometry, i.e. the same-indexed
// funcnodes must be in the same location, and the field must be
// defined.  Called by meshmenu.copyMesh.

void CSubProblem::acquire_field_data(Field &field, const CSubProblem *other) {
  for(FuncNodeIterator i=funcnode_iterator(); !i.end(); ++i) {
    for(int d=0; d<field.ndof(); ++d) {
      if(i.node()->hasField(field)) {
	field(i.node(), d)->value() = 
	  field(other->mesh->getFuncNode(i.count()), d)->value();
      }
    }
  }
}

bool CSubProblem::is_defined_field(const Field &field) const {
  return (field.index() < fielddata.size())
    && fielddata[field.index()].defined;
}

void CSubProblem::do_activate_field(Field &field) {
  if(!is_active_field(field)) {
    n_active_field++;
    FieldData &fdata = fielddata[field.index()];
    fdata.active = true;
    // The flag in the DegreeOfFreedom objects is *not* set here.
    // It's set only during the solution process (by
    // CSubProblem::mapdofeqs()).  That's because DegreeOfFreedoms are
    // shared among subproblems, but activity is a per-subproblem
    // trait.
  }
}

void CSubProblem::do_deactivate_field(Field &field) {
  if(is_active_field(field)) {
    n_active_field--;
    fielddata[field.index()].active = false;
  }
}

bool CSubProblem::is_active_field(const Field &field) const {
  return (field.index() < fielddata.size()) && fielddata[field.index()].active;
}

int CSubProblem::n_active_fields() const {
  return n_active_field;
}

// void CSubProblem::set_in_plane(const Field &field, bool planarity) {
//   if(field.index() >= fielddata.size()) {
//     fielddata.resize(field.index()+1);
//   }
//   fielddata[field.index()].in_plane = planarity;
// }

bool CSubProblem::in_plane(const Field &field) const {
  return mesh->in_plane(field);
//   return (field.index() < fielddata.size())
//     && fielddata[field.index()].in_plane;
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// Return a list of all the defined compound fields.  Compound fields
// are the ones that correspond directly to physical fields.

std::vector<CompoundField*>* CSubProblem::all_fields() const {
  std::vector<CompoundField*>* flist = new std::vector<CompoundField*>;
  std::vector<CompoundField*> &allfields = CompoundField::allcompoundfields();
  flist->reserve(fielddata.size());
  for(std::vector<CompoundField*>::size_type i=0; i<allfields.size(); i++) {
    // TODO PLASTICITY: This only picks up explicitly defined Fields.
    // Should it look at Node::fieldset instead?  Should there be
    // another function that retrieves implicit Fields that may not be
    // in CSubProblem::fielddata?
    if(is_defined_field(*allfields[i])) 
      flist->push_back(allfields[i]);
  }
  return flist;
}

// Likewise for Equation objects.
std::vector<Equation*>* CSubProblem::all_equations() const {
  std::vector<Equation*> *eqlist = new std::vector<Equation*>;
  std::vector<Equation*> &alleqns = Equation::all();
  eqlist->reserve(eqndata.size());
  for(std::vector<Equation*>::size_type i=0; i<alleqns.size(); i++) {
    if(is_active_equation(*alleqns[i]))
      eqlist->push_back(alleqns[i]);
  }
  return eqlist;
}

std::vector<Flux*>* CSubProblem::all_fluxes() const {
  std::vector<Flux*> *fluxlist = new std::vector<Flux*>;
  std::vector<Flux*> &allfluxen = Flux::allfluxes();
  for(std::vector<Flux*>::size_type i=0; i<allfluxen.size(); i++) {
    if(is_active_flux(*allfluxen[i]))
      fluxlist->push_back(allfluxen[i]);
  }
  return fluxlist;
}

std::vector<Flux*> CSubProblem::allFluxes() const {  // non-swigged version.
  std::vector<Flux*> fluxlist;
  std::vector<Flux*> &allfluxen = Flux::allfluxes();
  for(std::vector<Flux*>::size_type i=0; i<allfluxen.size(); i++) {
    if(is_active_flux(*allfluxen[i]))
      fluxlist.push_back(allfluxen[i]);
  }
  return fluxlist;
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

void CSubProblem::make_linear_system() {
//   Trace("CSubProblem::make_linear_system");
  Progress * localbar = getProgress();
  mesh->housekeeping();
  // std::cerr << "nodal equations =" << nodaleqn.size() << std::endl;
  // std::cerr << "dofs =" << dof.size() << std::endl;

  // Initialize the full linear system -- sets up all three matrices.
  // TODO DEVEL: Mindlessly clearing it here makes it work like the
  // old code.  In general, one wants to keep track of whether or not
  // one is evaluating at a new time, and only clear when you're sure
  // the time-step algorithm won't need the old data.  Or maybe the
  // time-stepper itself could keep the old data.
  linearsystem.clear();
  linearsystem.initialize(mesh->nodaleqn.size(), mesh->dof.size());
  if (kinetics_activated()) 
    linearsystem.set_kinetic();
  if (dynamics_activated())
    linearsystem.set_dynamic();

  for(ElementIterator ei=element_iterator(); !ei.end(); ++ei) {
    ei.element()->make_linear_system(this, linearsystem);
    localbar->set_progress(float(ei.count()+1)/float(ei.size()));
    localbar->set_message("creating linear system: " +
			  to_string(ei.count()+1) + "/" + to_string(ei.size()));
  }
//  std::cerr << "CSubProblem::make_linear_system:\n" << linearsystem.get_stiffness() << std::endl;
}

// Preliminary -- no housekeeping, just run the element's
// post-equilibration loop.
void CSubProblem::post_process() {
  //Thread control starts here
  Progress * localbar = getProgress(); //not a law firm
  //Thread control ends here
  for(ElementIterator ei=element_iterator(); !ei.end(); ++ei) {
    ei.element()->post_process(this);
    localbar->set_progress(float(ei.count()+1)/float(ei.size()));
    localbar->set_message("postprocessing: " + to_string(ei.count()+1) + "/"
			  + to_string(ei.size()));
  }
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// Routines used to pick out rows and columns of the stiffness matrix
// when extracting submatrices.

bool CSubProblem::solve_for_this_dof(int i) const {
  //dof list corresponds to unsymmetric system
  DegreeOfFreedom *dof = mesh->dof[i];
  return linearsystem.is_nonempty_col(i) && dof->isfree() && dof->active();
}

bool CSubProblem::dont_solve_for_this_dof(int i) const {
  //dof list corresponds to unsymmetric system
  DegreeOfFreedom *dof = mesh->dof[i];
  return linearsystem.is_nonempty_col(i) && (!dof->isfree() || !dof->active());
}

bool CSubProblem::solve_this_nodal_eqn(int j) const {
  //nodaleqn corresponds to unsymmetric system
  NodalEquation *ndq = mesh->nodaleqn[j];
  bool result = linearsystem.is_nonempty_row(j) && !ndq->is_dependent();
  return result;
}

bool CSubProblem::dont_solve_this_nodal_eqn(int j) const {
  return linearsystem.is_nonempty_row(j) && mesh->nodaleqn[j]->is_dependent();
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// Extract the submatrices of the stiffness matrix
// The conceptual subdivision is this:
//
//                 free      fixed
//                          |
//   independent      A     | C
//                          |
//                ----------+---     --- 
//     dependent      C'    | B
// 

// Rowmapper and Colmapper.
// Build a map of the rows i of the stiffness matrix for
// which f(i) is true, and return it.  Call with f(i)
// discriminating between independent (known-force) vs.
// dependent (unknown-force) equations for rows, or
// with f(i) discriminating between fixed and free
// DOFs for columns.
//
// Do this by manipulating the pre-existing mapping list.
void CSubProblem::rcmapper(bool (CSubProblem::*f)(int) const,
			   std::vector<int> &rcmap, int len) const
{
  //  Trace("FEMesh::rcmapper");
  int nrc = 0;			// # of rows or columns in the mapped submatrix
  rcmap.resize(len); 
  for(int i=0; i<len; i++) {
    if ((this->*f)(i)) {
      rcmap[i] = nrc++;
    }
    else {
      rcmap[i] = -1; // -1 is default for rows/columns which don't occur.
    }
  }
}

// Run the maps, creating the lists used to partition the stiffness
// matrix.

void CSubProblem::mapdofeqs() {
//   Trace("CSubProblem::mapdofeqs");

  // Set the "active" flag in the dofs.  This flag is used by
  // solve_for_this_dof and dont_solve_for_this_dof.  Setting the flag
  // here is a hack which will fail if more than one subproblem is
  // solved simultaneously.  But solving more than one subproblem
  // simultaneously is problematic anyway, if they share dofs.
  std::vector<Field*> &fields = Field::all();
  for(unsigned int i=0; i<fields.size(); i++) {
    Field &field = *fields[i];
    if(is_active_field(field)) {
      for(FuncNodeIterator ni=funcnode_iterator(); !ni.end(); ++ni) {
	if(ni.node()->hasField(field)) {
	  for(int i=0; i<field.ndof(); i++) {
	    field(ni.node(), i)->activate();
	  }
	} // if node has field
      }	// loop over nodes
    } // if field is active
    else {			// field is not active
      for(FuncNodeIterator ni=funcnode_iterator(); !ni.end(); ++ni) {
	if(ni.node()->hasField(field)) {
	  for(int i=0; i<field.ndof(); i++) {
	    field(ni.node(), i)->deactivate();
	  }
	} // if node has field
      }	// loop over nodes
    } // if field is not active
  } // loop over fields

  rcmapper(&CSubProblem::solve_this_nodal_eqn, indepeqnmap, 
	   linearsystem.nrows());
  // std::cerr << "  indepeqnmap= " << indepeqnmap << std::endl;
  rcmapper(&CSubProblem::dont_solve_this_nodal_eqn, depeqnmap, 
	   linearsystem.nrows());
  // std::cerr << "    depeqnmap= " << depeqnmap << std::endl;
  rcmapper(&CSubProblem::solve_for_this_dof, freedofmap, 
	   linearsystem.ncols()); 
  // std::cerr << "   freedofmap= " << freedofmap << std::endl;
  rcmapper(&CSubProblem::dont_solve_for_this_dof, fixeddofmap, 
	   linearsystem.ncols()); 
  // std::cerr << "  fixeddofmap= " << fixeddofmap << std::endl;

#ifdef HAVE_MPI_REDO
  //A note that might be useful: When the equations are made dependent
  //in bdycondition.py, the dependent i'th row of the symmetrized
  //matrix corresponds to the element of the vector nodaleqn[i]
  set_precombined_equation_mapping();
#endif
} 

// Undo BC invocation by resetting dofs and nodalequations to be free
// and independent.  Does not modify "active" flags.
void CSubProblem::resetdofeqs() {
  for(std::vector<DegreeOfFreedom*>::size_type i=0;i<mesh->dof.size();i++) 
    if(linearsystem.is_nonempty_col((int) i))
      mesh->dof[i]->free();
  for(std::vector<NodalEquation*>::size_type i=0;i<mesh->nodaleqn.size();i++)
    if(linearsystem.is_nonempty_row((int) i))
      mesh->nodaleqn[i]->independent();
}

// The cleanlists operation restores the row and column maps to 
// consistency after the application of a FloatBC.  The FloatBC's
// cause several rows or columns of the original stiffness matrix
// to be mapped (summed, actually) to a single row or column of
// the extracted 'A' matrix used in the actual linear-algebra solver.
// As a result, the range of the maps can develop holes -- cleanlists()
// closes those holes, ensuring that the range of the map is a 
// contiguous group of integers from 0 to the appropriate N.
void CSubProblem::cleanlists() {
  cleanmap(indepeqnmap);
  cleanmap(depeqnmap);
  cleanmap(freedofmap);
  cleanmap(fixeddofmap);
}

// This is a private routine that just does one map, called
// from FEMesh::cleanlists(). 
void CSubProblem::cleanmap(std::vector<int>& rcmap) {
  std::vector<int> histogram(rcmap.size(),0);
  std::vector<int> subtractor(rcmap.size(),0);

  // Make the histogram.
  for(std::vector<int>::size_type i=0;i<rcmap.size();i++) {
    if (rcmap[i]!=-1)
      histogram[ rcmap[i] ]++;
  }
  // To actually clean the list, the technique is this:
  // Move along the histogram, counting the number of zero
  // entries you see.  When you get to a nonzero entry, 
  // make an entry in the subtraction map -- the amount
  // to subtract from the value you're looking at is the
  // current zero count.  Then, make a pass through the
  // subtraction map and do the actual subtractions.
  int zcount = 0;
  for(std::vector<int>::size_type i=0 ; i<rcmap.size(); i++) {
    if(histogram[i]==0) {
      zcount++;
    }
    else {
      subtractor[i] = zcount;
    }
  }

  // Now do the subtraction.
  for(std::vector<int>::size_type i=0; i<rcmap.size(); i++) {
    if( rcmap[i] != -1 ) {
      rcmap[i] -= subtractor[ rcmap[i] ];
    }
  }
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

void CSubProblem::get_unknowns(VECTOR_D &x, const std::vector<int> &map) const {
  mesh->get_unknowns(x, map);
}

void CSubProblem::set_unknowns(const VECTOR_D &x, const std::vector<int> &map) {
  mesh->set_unknowns(x, map);
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// Extract the submatrices of the stiffness matrix.  Mapping arrays
// should be set up by the time you call this.

const SparseLinkMatrix<double> &CSubProblem::stiffness_matrix() const {
  if(!linearsystem.is_initialized()) {
    throw ErrProgrammingError("Stiffness matrix not computed!",
			      __FILE__, __LINE__);
  } 
  return linearsystem.get_stiffness();
}

const SparseLinkMatrix<double> &CSubProblem::conductivity_matrix() const {
  if(!linearsystem.is_kinetic()) {
    throw ErrProgrammingError("Conductivity matrix not computed!",
			      __FILE__, __LINE__);
  } 
  return linearsystem.get_kinetic();
}
const SparseLinkMatrix<double> &CSubProblem::mass_matrix() const {
  if(!linearsystem.is_dynamic()) {
    throw ErrProgrammingError("Mass matrix not computed!",
			      __FILE__, __LINE__);
  } 
  return linearsystem.get_dynamic();
}



// TODO DEVEL: These may want to move out of CSubProblem and into
// LinearSystem.
CSRmatrix CSubProblem::Amatrix() const {
//   Trace("CSubProblem::Amatrix");
  return linearsystem.stiffness_submatrix(indepeqnmap, freedofmap);
}

CSRmatrix CSubProblem::Bmatrix() const {
//   Trace("CSubProblem::Bmatrix");
  return linearsystem.stiffness_submatrix(depeqnmap, fixeddofmap);
}

CSRmatrix CSubProblem::Cmatrix() const {
//   Trace("CSubProblem::Cmatrix");
  return linearsystem.stiffness_submatrix(indepeqnmap, fixeddofmap);
}

CSRmatrix CSubProblem::Cpmatrix() const {
//   Trace("CSubProblem::Cpmatrix");
  return linearsystem.stiffness_submatrix(depeqnmap, freedofmap);
}

CSRmatrix CSubProblem::ConductivityMatrix() const {
  Trace("FEMesh::ConductivityMatrix");
  return linearsystem.kinetic_submatrix(indepeqnmap, freedofmap);
}

CSRmatrix CSubProblem::MassMatrix() const {
  Trace("FEMesh::MassMatrix");
  return linearsystem.dynamic_submatrix(indepeqnmap, freedofmap);
}

// Return the "blocking" data -- this is a map which is -1 if the
// corresponding row/column is in-plane, and is the node number if the
// corresponding row/column is out-of-plane.  Used by the block
// preconditioner, and by the Kinetic time stepping code.

// TODO: It might be more clever if these returned pointers/references
// to objects in either the subproblem or stiffness matrix classes.
std::vector<int> CSubProblem::A_rowblocks() const {
  return linearsystem.rowgroups(indepeqnmap);
}

std::vector<int> CSubProblem::A_colblocks() const {
  return linearsystem.colgroups(freedofmap);
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

void CSubProblem::get_rhs(VECTOR_D &rhs, const std::vector<int> &map) const {
//   Trace("FEMesh::get_rhs");
  VECTOR_D &body_rhs = linearsystem.get_body_rhs();
  VECTOR_D &boundary_rhs = linearsystem.get_boundary_rhs();
  for(std::vector<NodalEquation*>::size_type i=0; i<mesh->nodaleqn.size(); i++)
    {
      if(map[i] != -1) {
	rhs[map[i]] += body_rhs[i] + boundary_rhs[i];
	// Merge nodal equation values with boundary RHS contribs.
      }
    }
}

VECTOR_D &CSubProblem::boundary_rhs() const {
  if(!linearsystem.is_initialized()) {
    throw ErrProgrammingError("Boundary right hand side not available!",
			      __FILE__, __LINE__);
  }
  return linearsystem.get_boundary_rhs();
}

void CSubProblem::clear_brhs() {
  linearsystem.clear_brhs();
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// Routines for symmetrizing the matrix (or trying to), using
// conjugacy data.

// In the case of inconsistent conjugacy data, build a trivial
// equation map.
void CSubProblem::set_trivial_equation_mapping() {
  if (global_equation_map) {
    delete global_equation_map;
  }
  global_equation_map = new std::vector<int>(mesh->nodaleqn.size());
  for(unsigned int i=0; i<global_equation_map->size(); ++i) {
    (*global_equation_map)[i]=i;
  }
}

// See if a local conjugacy map has already been constructed for the
// given fieldset and nodeset, and construct a new one if necessary.
// (In this context, "local" means "local to a Node".)  The local map
// is a vector of ints.  For the i^th equation in a FuncNode's list of
// NodalEquations, localmap[i] is the index of the conjugate degree of
// freedom in the Node's list of DegreeOfFreedoms.  If there is no
// conjugate degree of freedom, localmap[i] is -1.

static std::vector<int> &getLocalMap_(FuncNode::FieldSet &fieldset,
				      FuncNode::EquationSet &eqnset,
				      LocalMapDict &localmaps,
				      CSubProblem *subproblem,
				      const std::vector<CConjugatePair*> &pairs)
{
  FEPair key(fieldset, eqnset);
  LocalMapDict::iterator i = localmaps.find(key);
  if(i == localmaps.end()) {
    // create a new map
    std::vector<int> lmap(eqnset.dimsum(), -1); // rub it and a gneie appears
    for(std::vector<CConjugatePair*>::const_iterator p=pairs.begin();
	p < pairs.end(); ++p)
      {
	const Field &field = *(*p)->get_field();
	const Equation &eqn = *(*p)->get_equation();
	if(subproblem->is_active_field(field) &&
	   subproblem->is_active_equation(eqn))
	  {
	    const FieldIndex &equationcomp = *(*p)->get_equation_component();
	    const FieldIndex &fieldcomp = *(*p)->get_field_component();
	    int eindx = eqnset.offset(&eqn) + equationcomp.integer();
	    int findx = fieldset.offset(&field) + fieldcomp.integer();
	    lmap[eindx] = findx;
	  }
      }
    localmaps.insert(LocalMapDict::value_type(key, lmap)); // copies lmap
    return localmaps[key];	// returns reference to copy in localmaps
  }
  else {
    return (*i).second;		// reuse an existing map
  }
}


void CSubProblem::set_equation_mapping(
			  const std::vector<CConjugatePair*> *conjugacylist)
{
  // Build a mapping array such that, when you traverse
  // FEMesh::nodaleqn[map[i]] for i from 0 to FEMesh::nodaleqn.size(),
  // the nodal equations come up in the same order as their conjugate
  // degrees of freedom do when you index that array (FEMesh::dof)
  // directly.  Store this map in the global_equation_map variable.
  // Using this ordering at stiffness-matrix-construction time causes
  // the resulting stiffness matrix to be as symmetric as possible.

  // Any equations for which conjugate DOFs do not exist come at the
  // tail of this list in no particular order.

  // For any given i, and in a given node's lists, the cols[i]-th
  // degree of freedom in the node is conjugate to the rows[i]-th
  // nodal equation in the node, with i ranging over the number of
  // active equations or degrees of freedom -- the number of *active*
  // eqns/dofs must match, but there may be inactive dofs.  There are
  // never inactive equations.

  if (global_equation_map) {
    delete global_equation_map;
  }
  global_equation_map = new std::vector<int>(mesh->nodaleqn.size(), -1);
  
  // Make sure the lists are contiguous.
  mesh->housekeeping();

#ifdef HAVE_MPI_REDO
  ResetSharesData();
#endif

  LocalMapDict localmaps;
  int max_target_row = -1;
  for(FuncNodeIterator iter=funcnode_iterator(); !iter.end(); ++iter) {
    FuncNode *node = iter.node();

    std::vector<NodalEquation*> &eqnlist = node->eqnlist;
    // Get the localmap, which indicates the conjugacy relations
    // between nodal equations and degrees of freedom in the Node.
    std::vector<int> &localmap = getLocalMap_(node->fieldset,
					      node->equationset,
					      localmaps, this, *conjugacylist);
    for(std::vector<NodalEquation*>::size_type i=0; i<eqnlist.size(); ++i) {
      // Only do stuff if this mapping makes sense.
      if (localmap[i]!=-1) {
	NodalEquation* neq = eqnlist[i];
	int source_row = neq->ndq_index(); // Index in master stiffness mtx

	DegreeOfFreedom *dof = (node->doflist)[localmap[i]];
	int target_row = dof->dofindex();  // Also in master stiffness mtx
	
	(*global_equation_map)[source_row] = target_row;

#ifdef HAVE_MPI_REDO
	if(m_eqnnodemap.find(neq)!=m_eqnnodemap.end())
	  m_eqnmatrixindexmap[neq]=target_row;
#endif

	if(target_row > max_target_row)
	  max_target_row = target_row;
      }
    }
  }
  
  // There may be unmapped equations due to incomplete conjugacy data.
  // Also, there may be equations for which there are no degrees of
  // freedom because they're going to be dependent equations.  In
  // either case, these "extra" equations will not have been mapped
  // anywhere.  Make a second pass through the list, and assign any
  // unassigned equations an incremental index, starting with
  // max_target_row.  Bounds are not an issue, because of the
  // range-compression step which follows this block.
  for(FuncNodeIterator iter=funcnode_iterator(); !iter.end(); ++iter) {
    FuncNode *node = iter.node();
    for(std::vector<NodalEquation*>::size_type i=0; 
	i<(node->eqnlist).size(); ++i) {
      NodalEquation *neq = (node->eqnlist)[i];
      int source_index = neq->ndq_index(); 
      if ( (*global_equation_map)[source_index] == -1 ) {
	max_target_row++;
	(*global_equation_map)[source_index]=max_target_row;
#ifdef HAVE_MPI_REDO
	if(m_eqnnodemap.find(neq)!=m_eqnnodemap.end())
	  m_eqnmatrixindexmap[neq]=max_target_row;
#endif
      }
    }
  }

  // Now fix the range of the global_equation_map -- the master
  // stiffness matrix can be wider than it is tall, so some of the
  // equations could be out of range.  Fix this by compressing the
  // range but preserving the order.

  // First, build a map indexed by the *range* elements of the map,
  // but preserving the domain information.  This is the "inverse" of
  // the forward-map.
  std::map<int,int> back_map;
  for(std::vector<int>::size_type i=0; i<global_equation_map->size(); ++i) {
    back_map[(*global_equation_map)[i]]=i;
  }
  // Then go through the backward map in range-of-forward-map
  // (domain-of-backward-map, key of std::map structure) order, and
  // for each element of the domain of the forward map (range of the
  // backward map, value of std::map structure), assign consecutive
  // integers to the range of the forward map.
#ifdef HAVE_MPI_REDO
  std::vector<NodalEquation*> visitedEqns;
#endif
  int key_count=0;
  // Note: This loop traverses back_map sorted in the order
  // according to the keys (bmi->first).
  for(std::map<int,int>::const_iterator bmi=back_map.begin();
      bmi!=back_map.end(); ++bmi)
    {
      (*global_equation_map)[(*bmi).second]=key_count;
#ifdef HAVE_MPI_REDO
      // Adjust the values for our map again.  Search for the
      // NodalEquation with value bmi->first in the map
      // m_eqnmatrixindexmap
      // TODO: Speed this up later!
      for(std::map<NodalEquation*,int>::iterator it=m_eqnmatrixindexmap.begin();
	  it!=m_eqnmatrixindexmap.end(); it++)
	{
	  if(find(visitedEqns.begin(),visitedEqns.end(),it->first)==visitedEqns.end() &&
	     it->second==bmi->first)
	    {
	      it->second=key_count;
	      visitedEqns.push_back(it->first);
	      break;
	    }
	}
#endif
      key_count++;
    }

#ifdef HAVE_MPI_REDO
  // From eqnmatrixindexmap, give CNodeShareInfo the information about where
  //      in the local stiffness matrix the equation/row can be found
  for(std::map<NodalEquation*,int>::iterator it=m_eqnmatrixindexmap.begin();
      it!=m_eqnmatrixindexmap.end(); it++)
    {
      NodalEquation* pne=it->first;
      //Look for the FuncNode associated with this NodalEqn
      FuncNode* pfn=m_eqnnodemap[pne];
      //Get sharing information for FuncNode
      CNodeShareInfo* pns=m_nodesharemap[pfn];
      //Get index of NodalEqn in FuncNode->eqnlist
      int localindex=(find(pfn->eqnlist.begin(),pfn->eqnlist.end(),pne)-
		      pfn->eqnlist.begin());
      // I assume that the local index of a field/DoF/NodalEquation in
      // a node is the same for all nodes, (TODO MPI: NOT TRUE ANYMORE!)
      // since they are added the same way for each FuncNode.  The
      // independent,dependent(for equations) and fixed,free (DoFs)
      // status should also be the same for each shared node.
      pns->localeqnindexlist.push_back(localindex);
      pns->symmatrixeqnindexlist.push_back(it->second);
      // it->second is the row number in the stiffness matrix. We
      // shall put this row later into its right place in the combined
      // linear system
    }
  // Give CNodeShareInfo the information about where in the local
  // stiffness matrix the dof/column can be found
  for(std::map<DegreeOfFreedom*, FuncNode*>::iterator it=m_dofnodemap.begin();
      it!=m_dofnodemap.end();it++)
    {
      DegreeOfFreedom* pdof=it->first;
      //FuncNode associated with this DoF
      FuncNode* pfn=it->second;
      //Get sharing information for FuncNode
      CNodeShareInfo* pns=m_nodesharemap[pfn];
      //Get index of DoF in FuncNode->doflist
      int localindex=(find(pfn->doflist.begin(),pfn->doflist.end(),pdof)-
		      pfn->doflist.begin());
      pns->localdofindexlist.push_back(localindex);
      //Get index of DoF in the mesh->dof vector
      int mtxindex=(find(this->dof.begin(),this->dof.end(),pdof)-this->dof.begin());
      pns->symmatrixdofindexlist.push_back(mtxindex);
    }
  //RCL testing
  if(0)
    {
      for(std::map<FuncNode*,CNodeShareInfo*>::iterator it=m_nodesharemap.begin();
	  it!=m_nodesharemap.end();it++)
	{
	  CNodeShareInfo* pns=it->second;
	  std::cerr << "****** Inherited index ****** " << pns->inheritedindex << std::endl;
	  for(std::vector<int>::size_type i=0;i<pns->localeqnindexlist.size();i++)
	    {
	      std::cerr << "NodalEqn " << pns->localeqnindexlist[i] << "," <<
		pns->symmatrixeqnindexlist[i] << std::endl;
	    }
	  for(std::vector<int>::size_type i=0;i<pns->localdofindexlist.size();i++)
	    {
	      std::cerr << "DoF " << pns->localdofindexlist[i] << "," <<
		pns->symmatrixdofindexlist[i] << std::endl;
	    }
	}
    }
#endif // HAVE_MPI

} // set_equation_mapping

void CSubProblem::clear_symmetricity_maps()
{
  //destroy the symmetricity maps after using them
  delete global_equation_map;
  global_equation_map=0;
}

// Having remapped the equation array, we need a helper function so
// that boundary conditions can set the appropriate equations to be
// dependent.  

const NodalEquation &CSubProblem::map_forward_equation(const Equation &equation,
						       const FuncNode &node,
						       const int &eqcomp) const
{
  int target_index = map_forward_index(equation, node, eqcomp);
  return *(mesh->nodaleqn[target_index]);
}

int CSubProblem::map_forward_index(const Equation & equation, 
				   const FuncNode &node, const int &eqcomp)
  const 
{
  // Reads the global equation map -- see set_equation_mapping for details.
  int original_row = equation.nodaleqn(node, eqcomp)->ndq_index();
  return (*global_equation_map)[original_row];
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// dump_dof and dump_eqn are commented out because they require const
// iterators,which I'm too lazy to write just now.

// void CSubProblem::dump_dof() const {
//   for(FuncNodeIterator i=funcnode_iterator(); !i.end(); ++i) {
//     const FuncNode &node = *i.node();
//     std::cout << "node = " << node << std::endl;
//     for(std::vector<Field*>::size_type f=0; f<Field::all().size(); f++) {
//       const Field &field = *Field::all()[f];
//       if(field.is_defined(this)) {
// 	for(int j=0; j<field.ndof(); j++) {
// 	  std::cout << "\t" << field.name() << "[" << j << "] dof="
// 	       << *field(node, j) << std::endl;
// 	}
//       }
//     }
//     for(std::vector<Equation*>::size_type e=0;e<Equation::all().size();e++){
//       const Equation &eqn = *Equation::all()[e];
//       if(eqn.is_active(this)) {
// 	for(int j=0; j<eqn.dim(); j++) {
// 	  std::cout << "\t" << eqn.name() << "[" << j << "] nodaleqn="
// 	       << *eqn.nodaleqn(node, j) << std::endl;
// 	}
//       }
//     }
//   }
// }


// void CSubProblem::dump_eqn() const {
//   for(FuncNodeIterator i=funcnode_iterator(); !i.end(); ++i) {
//     const FuncNode &node = *i.node();
//     std::cout << "node = " << node << std::endl;
//     for(std::vector<Equation*>::size_type e=0;e<Equation::all().size();e++){
//       const Equation &eqn = *Equation::all()[e];
//       if(eqn.is_active(this)) {
// 	for(int j=0; j<eqn.dim(); j++) {
// 	  std::cout << "\t" << eqn.name() << "[" << j << "] nodaleqn="
// 	       << *eqn.nodaleqn(node, j) << std::endl;
// 	}
//       }
//     }
//   }
// }

int CSubProblem::check_symmetricity(const CSRmatrix &A) {
  for (int i=0; i<A.nrows(); i++)
    for (int j=0; j<A.ncols(); j++)
      if (A(i,j) != A(j,i)) {
	std::cerr << "(i,	j)= (" << i <<"," << j << ")" << std::endl;
	std::cerr << "A(i,j)=" << A(i,j) << " A(j,i)=" << A(j,i) << std::endl;
	return 0;
      }
  return 1;
}


void CSubProblem::requirePrecompute() {
  precomputeLock.acquire();
  precomputeRequired = true;
  precomputeLock.release();
}

