// -*- C++ -*-
// $RCSfile: femesh.h,v $
// $Revision: 1.85 $
// $Author: rlua $
// $Date: 2007/02/14 22:37:25 $


/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

#include <oofconfig.h>

#ifndef FEMESH_H
#define FEMESH_H

class FEMesh;

#include <Python.h>

#include "engine/field.h"
#include "engine/equation.h"
#include "engine/fieldeqnlist.h"
#ifdef HAVE_MPI
#include "common/mpitools.h"	// TODO: is this needed here?
#endif
#include <string>
#include <vector>
#include <map>

class CMicrostructure;
class Coord;
class DegreeOfFreedom;
class Element;
class ElementIterator;
class Equation;
class Field;
class FuncNode;
class FuncNodeIterator;
class MasterCoord;
class NodalEquation;
class NodalFluxes;
class NodalSCPatches;
class Node;
class NodeIterator;
class RWLock;


//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

#ifdef HAVE_MPI
class CNodeShareInfo
{
public:
  CNodeShareInfo(std::vector<int>* procs,
		 std::vector<int>* indices, int index);
  ~CNodeShareInfo();
   // index inherited from the skeleton nodes sharing information
  int inheritedindex;		
  //the rank of the process
  int localprocrank;
  // Does this process own the node? It does if the process is the
  // lowest ranked among those that share this node.
  bool owns;
  // list of processes (rank numbers) that share this node:
  std::vector<int> remoteproclist;
  // list of indices of this node corresponding to the other processes
  // listed in remoteproclist:
  std::vector<int> remoteindexlist;
  // The following four vectors should have the same length.  These
  // get filled up in set_equation_mapping.
  // list of indices of the free DoFs within this node:
  std::vector<int> localdofindexlist;
  // list of indices of the independent NodalEqns within this node:
  std::vector<int> localeqnindexlist;
  // list of indices into the (probably) symmetrized stiffness matrix
  // //of the free DoFs:
  std::vector<int> symmatrixdofindexlist;
  // list of indices into the (probably) symmetrized stiffness matrix
  // of the independent NodalEqns:
  std::vector<int> symmatrixeqnindexlist;
};
// Then construct a mapping of the indices in the symmatrixdofindexlist to the
//  combined/global/"uber" linear system.
// Also need a list of the localdofsizes [dofsize0,dofsize1,...]

#endif // HAVE_MPI

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

class FEMesh {    // : public PythonExportable<FEMesh> {
  // FEMesh has to be derived from PythonExportable so that it can be
  // used as an argument in switchboard messages.
private:
  CMicrostructure * microstructure;
  RWLock *rwlock;
public:
  FEMesh(CMicrostructure *);
  virtual ~FEMesh();
  CMicrostructure *get_microstructure() const { return microstructure; }
  virtual const std::string &classname() const;
  virtual const std::string &modulename() const;

  ElementIterator element_iterator() const;
  NodeIterator node_iterator() const;
  FuncNodeIterator funcnode_iterator() const;

  Node *newMapNode(const Coord&); // the only way to make a Node
  FuncNode *newFuncNode(const Coord&); // the only way to make a FuncNode

#ifdef HAVE_MPI
  // These carry the extra sharing information derived from the
  // parallel skeleton construction of Haan
  FuncNode *newFuncNode_shares(const Coord&,std::vector<int> * procs,
			       std::vector<int> * remoteindices, int index);
  // Map of inherited index to shared FuncNodes
  std::map<int,FuncNode*> m_indexnodemap;
  std::map<FuncNode*,CNodeShareInfo*> m_nodesharemap;
  // These maps are keyed by DoFs and NodalEqns.
  // These link to the sharing information stored in the FuncNode
  // and in the two maps above. The intention is to add to the map only
  // those DoFs and NodalEqns that are associated
  // with a node that is shared between two or more processes.
  // In principle, one can search the doflist or eqnlist of the FuncNode's
  // to get the right FuncNode, but this is slower. (This may be temporary)
  std::map<DegreeOfFreedom*, FuncNode*> m_dofnodemap;
  std::map<NodalEquation*, FuncNode*> m_eqnnodemap;
#if 0//Redo
  // Map of a nodal equation to the row of the (probably symmetrized)
  // matrix that it corresponds to.
  std::map<NodalEquation*,int> m_eqnmatrixindexmap;
  //
  std::map<DegreeOfFreedom*,int> m_dofmatrixindexmap;//Not used (yet?)
  // These are arrays corresponding to the freedofmap and indepeqnmap
  // but with the shared non-owned rows and columns indicated.  At the
  // end of the day, these should map the indices of the local
  // stiffness matrix onto the global stiffness matrix
  std::vector<int> m_precombined_freedofmap;
  std::vector<int> m_precombined_indepeqnmap;
  // These are less than or equal to m_precombined_freedofmap.size()
  // (not counting the -1)
  int m_precombined_dofsize;
  int m_precombined_eqnsize;
  void set_precombined_equation_mapping();
  // This member stores the cumulative sum of pdofsizelist
  // (dofsize0+dofsize1+...)
  int m_combinedmatrixdim;
  // This function for displaying solution at the FuncNodes
  void NodalPositionSolution(double* temp_array);
  //
  void ResetSharesData();
#endif
#endif	// HAVE_MPI

  void reserveFuncNodes(int n);	// reserve space for n func nodes
  void reserveMapNodes(int n);	// reserve space for n map nodes
  void AddElement(Element*);
  void reserveElements(int n);

  // Caution: NodeIterator::index is not necessarily the same as
  // node.index().  The argument to FEMesh::getNode() is the
  // NodeIterator::index.
  Node *getNode(int) const;
  FuncNode *getFuncNode(int) const;

  // Temporary function for finding the closest node. TODO: Write a better one.
  Node *closestNode(const double x, const double y);

  DegreeOfFreedom *CreateDoF(double x=0); // only way to make a DoF
  void removeDoF(DegreeOfFreedom*);

  // NodalEquations represent one component of one Equation at a Node.
  // They are to the rows of the stiffness matrix what the
  // DegreeOfFreedoms are to the columns.
  NodalEquation *CreateNodalEqn();
  void removeNodalEqn(NodalEquation*);

  Element *getElement(int i) const;

  void refreshMaterials(PyObject *skeleton);

  int nnodes() const;
  int nelements() const;

  // Is a field in-plane on this mesh?
  bool in_plane(const Field &field) const;
  void set_in_plane(const Field &field, bool);
  std::vector<std::string> *getFieldSetByID(int) const;	// returns new obj

  // API for setting/referring to the read-write lock.  Set_rwlock
  // should be called exactly once when the femesh is inserted into a
  // mesh context object.
  void set_rwlock(RWLock *rw) { rwlock = rw; };
  inline RWLock *get_rwlock() { return rwlock;};

private:
  // These lists can be accessed through the MeshIterators.
  std::vector<FuncNode*> funcnode; // nodes at which dofs are defined
  std::vector<Node*> mapnode;	// nodes only used for geometry mapping
  std::vector<Element*> element;
  int ncount;			// node counter, for assigning ids

  // Master lists of degrees of freedom and nodal equations
  std::vector<DegreeOfFreedom*> dof;
  std::vector<NodalEquation*> nodaleqn;
  void housekeeping();		     // do garbage collection on the lists
  bool dof_list_needs_cleaning;
  bool nodaleqn_list_needs_cleaning; // do the lists need to be cleaned up?

  // Dictionary of wrappers for Field and Equation lists stored at
  // Nodes.  See fieldeqnlist.h.
  FEWrapper<Field>::AllWrappers fieldwrappers;
  FEWrapper<Equation>::AllWrappers equationwrappers;

  std::vector<bool> in_plane_field;

  void clean_nodaleqn();
  void clean_doflist();

public:
  // Routines for extracting parts of the rhs and solution vectors
  void get_unknowns(VECTOR_D&, const std::vector<int>&) const;
  void set_unknowns(const VECTOR_D&, const std::vector<int>&);

private:
  friend class Equation::FindAllEquationWrappers;
  friend class Field::FindAllFieldWrappers;
  friend class MeshElementIterator;
  friend class MeshFuncNodeIterator;
  friend class MeshNodeIterator;
  friend class Node;
  friend class CSubProblem;

// Adaptive Mesh Refinement stuff.  TODO: Move to SubProblem.
public:
  // TODO: encapsulate all of the ZZ error estimation stuff, so that
  // we can switch estimators.

  // create & add a new CSCPatch pointer
  void init_scpatches(const std::vector<int>*);
  void add_scpatch(const int, const Material*, 
		   const int,
		   const std::vector<int>*,
		   const std::vector<int>*,
		   const int);
  // getting elements & nodes from the patch
  std::vector<int> *get_elements_from_patch(const int, const Material*);
  std::vector<int> *get_nodes_from_patch(const int, const Material*);
  // recovering flux(es)
  void init_nodalfluxes();
  void recover_fluxes();
  // adding recovered flux
  void add_this_flux(const Material*, const Flux*,
		     const Node*, VECTOR_D*);
  // recovered flux at a given point
  VECTOR_D *get_recovered_flux(const Flux*, const Element*,
			       const MasterCoord&);
  // reporting recovered fluxes at a given point -- debug purpose
  void report_recovered_fluxes(const Element*, const Coord*);
  // estimating error
  double zz_L2_estimate(const Element*, const Flux*);
  void zz_L2_estimate_sub(const Element*, const Flux*,
			  const int&, double&, double&,
			  const MasterCoord&, const double&);
  std::vector<double> *zz_L2_weights(const Flux*,
				     const double&, const double&);
  void zz_L2_weights_sub(const Element*, const Flux*,
			  const int&, double&,
			  const MasterCoord&, const double&);
  void setDefaultSubProblem(CSubProblem*);
private:
  // Storage for CSCPatch's (keyed an assembly node)
  // NodalSCPatches contains as many CSCPatches as no. of Materials
  // at the assembly node.
  std::map<const int, NodalSCPatches*> scpatches;
  // Storage for SCpatch Recovered Fluxes
  std::map<const int, NodalFluxes*> recovered_fluxes;
  // The default subproblem is passed in so that the above routines
  // can use it.  This is a temporary hack, only to be used while AMR
  // is being done in the FEMesh instead of the the CSubProblem.
  CSubProblem *defaultSubProblem;

};				// FEMesh

#endif // FEMESH_H
