// -*- C++ -*-
// $RCSfile: field.h,v $
// $Revision: 1.34 $
// $Author: langer $
// $Date: 2006/12/19 20:29:44 $

/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

#ifndef FIELD_H
#define FIELD_H

class CompoundField;
class Field;
class FieldIndex;
class IteratorP;
class ScalarField;
class ScalarFieldBase;
class TwoVectorField;
class TwoVectorFieldBase;

#include "common/coord.h"
#include "freedom.h"
#include "common/identification.h"
#include "common/pythonexportable.h"
#include "engine/fieldeqnlist.h"
#include "engine/fieldindex.h"
#include "engine/indextypes.h"
#include "engine/matvec.h"
#include <iostream>
#include <string>
#include <vector>

class CSubProblem;
class ElementFuncNodeIterator;
class FEMesh;
class FuncNode;
class OutputValue;
class Property;

// There is one Field for each physical field.  There should be only
// one instance of each derived Field class.  These are just
// bookkeeping classes, and don't actually store the field values at
// each node. That's done by DegreeOfFreedom. So there's no problem
// with creating fields that aren't used. Thus, every possible field
// is constructed (in problems.C), and they're never deleted.

// Fields are *active* if they're being solved for.
// Fields are *defined* if the nodes contain values for them.

// Because a Field may be defined, or active, or in-plane, on one
// CSubProblem, but not on another CSubProblem, much of the data for a Field is
// stored in a CSubProblem::FieldData class inside the CSubProblem class.  This
// is meant to be completely transparent, but it does mean that a
// CSubProblem pointer has to be passed to some Field functions.

// Access to degrees of freedom at Nodes (specifically, FuncNodes) is
// done via routines like DegreeOfFreedom *Field::operator()(const
// FuncNode&), but these aren't defined in the base class, because
// they can act very differently depending on the type of
// field. (Since the Fields are global variables, there's no need to
// use virtual functions -- any Property needing to get the value of a
// Field has access to the derived class functions.)

// ARRGH! That's not quite right! CompoundField::out_of_plane()
// returns a pointer to a Field (base class). So we need at least one
// way of accessing DegreeOfFreedom objects through the base
// class. Therefore the function
//    DegreeOfFreedom *Field::operator()(const FuncNode&, int component)
// is virtual, even though the component argument doesn't make sense
// for some fields.

// From the point of view of the matrix construction, the fields
// themselves either have in-plane or out-of-plane degrees of freedom
// -- this status is stored in the in_plane_ data member, and can be
// queried by the in_plane_part function.  Fields are in_plane by
// default, and can be set out-of-plane by calling the set_oop
// function.  NOTE that this is different than the value returned by
// CompoundField::in_plane(FEMesh*), which indicates whether or not a
// compound field has an out-of-plane part on a particular mesh.  For
// example, the out-of-plane part of the temperature field, dT/dz,
// always return in_plane_part=0, but the compound Temperature field
// can return in_plane=0 or 1, depending on the problem being solved.

class Field : public IdentifiedObject
{
private:
  const std::string name_;
  const unsigned int index_;
  bool in_plane_;
protected:
  const int dim;
public:
  Field(const std::string &, int);

  const std::string &name() const;
  unsigned int index() const { return index_; }

  static Field *getField(const std::string &name);

  // where a given component lives in the dof lists in a Node
  int localindex(const FuncNode*, const FieldIndex &component) const;

  int ndof() const { return dim; } // number of degrees of freedom
  
  // See comment above about in_plane() and in_plane_part().
  void set_oop() { in_plane_ = false; }
  bool in_plane_part() const { return in_plane_; }
  
  virtual void activate(CSubProblem*);
  virtual void deactivate(CSubProblem*);
  bool is_active(const CSubProblem*) const;
  
  virtual void define(CSubProblem*);
  virtual void undefine(CSubProblem*);
  bool is_defined(const CSubProblem*) const;

  virtual void registerProperty(Property*) const;

  // First, a way to initialize just with a function pointer that
  // takes a position and a degree-of-freedom index
  virtual void initialize(CSubProblem*, double (*f)(const Coord &pos, int dof));

  // Second, a way to initialize with a function pointer that takes
  // a position, a DOF index, and a generic pointer-to-void that
  // can contain some state information, as well as the pointer-to-void
  // itself. This is to be able to initialize from Python in a generic
  // way to keep the python dependent code independent from this
  // implementation, but not have to use global or static variables.
  virtual void initialize(CSubProblem*,
			  double (*f)(const Coord &pos, int dof, void *p2v),
			  void *p2v);

  // all() needs to be a function, because it is used during the
  // construction of global Field objects.  Making it a function
  // guarantees that the vector that it returns is constructed before
  // it's needed.
  static std::vector<Field*> &all();

  double value(const FuncNode*, int component) const;
  double value(const ElementFuncNodeIterator&, int component)
    const;
  
  virtual DegreeOfFreedom *operator()(const FuncNode*, int component) const = 0;
  DegreeOfFreedom *operator()(const FuncNode &n, int component) const
  {
    return operator()(&n, component);
  }
  virtual DegreeOfFreedom *operator()(const ElementFuncNodeIterator&,
				      int component)
    const = 0;
  DegreeOfFreedom *operator()(const ElementFuncNodeIterator &n,
			      const IteratorP &i) const
  {
    return operator()(n, i.integer());
  }

  virtual OutputValue newOutputValue() const = 0;

  virtual OutputValue output(const ElementFuncNodeIterator&) const = 0;
  virtual OutputValue output(const FuncNode&) const = 0;
  virtual void setValueFromOutputValue(const FuncNode&, const OutputValue*) = 0;

  friend bool operator==(const Field &f1, const Field &f2) {
    return f1.index() == f2.index();
  }

  virtual IteratorP iterator(Planarity) const = 0;
  virtual IndexP componenttype() const = 0;
  virtual IndexP getIndex(const std::string&) const = 0;

  // Stuff required by fieldeqnlist.h templates, which handle Field
  // data that varies from Node to Node, such as specifying which
  // fields are defined and where they live in the Node's list of
  // values. These classes must be typedef'd so that they can be
  // located by the templates.
  class FieldData : public FieldEqnData {
  public:
    FieldData() : active(false) {}
    bool active;		// is a Field active at this node?
  };
  typedef FieldData FEData;
  // FindAllFieldWrappers is an object that fetches a Mesh's
  // dictionary of Field data.
  class FindAllFieldWrappers {
  private:
    FEMesh *mesh;
  public:
    FindAllFieldWrappers(FEMesh *mesh) : mesh(mesh) {}
    // Using the typedef FEWrapper<Field>::AllWrappers here doesn't
    // work.  The compiler can't untangle the templates.  Is this a
    // compiler bug, or just due to a failure to define everything
    // first?
    std::map<const std::vector<FieldData>*, FEWrapper<Field>*,
	     ltFEvector<Field> >
    &operator()();
  };
  typedef FindAllFieldWrappers GetWrappers;


protected:
  virtual ~Field();
};

bool operator<(const Field::FieldData&, const Field::FieldData&);

std::ostream &operator<<(std::ostream &, const Field&);

// Compound fields have two parts: the in-plane part, and the
// out-of-plane part.  The Fields that the rest of the program uses
// are derived from the Compound Fields (eg displacement,
// temperature).  CompoundField needs to be derived from Field to
// override the virtual functions for defining and activating the
// field.  Since the in- and out-of-plane fields also need to be
// derived from Field, we have to use virtual inheritance.

class CompoundField : public virtual Field {
private:
  Field * const zfield_;	// the out-of-plane field
  int cfield_indx;
protected:
  CompoundField(const std::string &name, int dim, Field *follies);
  virtual ~CompoundField();
public:
  Field *out_of_plane() const {
    return zfield_;		// the zfield follies?
  }
//   void set_in_plane(FEMesh*, bool);
  bool in_plane(const FEMesh*) const;
  bool in_plane(const CSubProblem*) const;
  virtual void define(CSubProblem*);
  virtual void undefine(CSubProblem*);
  virtual void activate(CSubProblem*);
  virtual void deactivate(CSubProblem*);
  virtual void registerProperty(Property*) const;
  static std::vector<CompoundField*> &allcompoundfields();
//   // For identification in python...
  virtual const std::string &classname() const = 0;
};

//-\\-//-\\-//-\\-//-\\-//-\\-//-\\-//-\\-//-\\-//-\\-//-\\-//-\\-//

// Different varieties of Fields

class ScalarFieldBase : public virtual Field {
public:
  ScalarFieldBase(const std::string &name) : Field(name, 1) {}
  virtual ~ScalarFieldBase() {}
  DegreeOfFreedom *operator()(const FuncNode*) const;
  DegreeOfFreedom *operator()(const FuncNode &n) const {
    return operator()(&n);
  }
  DegreeOfFreedom *operator()(const ElementFuncNodeIterator&) const;
  virtual DegreeOfFreedom *operator()(const FuncNode*, int) const;
  virtual DegreeOfFreedom *operator()(const FuncNode &n, int i) const {
    return operator()(&n, i);
  }
  virtual DegreeOfFreedom *operator()(const ElementFuncNodeIterator&, int)
    const;
  virtual OutputValue newOutputValue() const;
  virtual OutputValue output(const ElementFuncNodeIterator&) const;
  virtual OutputValue output(const FuncNode&) const;
  virtual void setValueFromOutputValue(const FuncNode&, const OutputValue*);

  virtual IteratorP iterator(Planarity=ALL_INDICES/*irrelevant*/) const;
  virtual IndexP componenttype() const;
  virtual IndexP getIndex(const std::string&) const;
};

// Since CompoundField and ScalarFieldBase are both derived from
// Field, the inheritance must be virtual.
class ScalarField : public ScalarFieldBase, public CompoundField {
 private:
  static const std::string classname_;
public:
  ScalarField(const std::string &name);
  virtual ~ScalarField() {}
  virtual const std::string &classname() const { return classname_; }
};

//------------------------

// This class allows TwoVectorFields to be used like Coords. There are
// no coordinate or DegreeOfFreedom extraction operators. If you want
// access to the individual degrees of freedom, use field(node, i)
// instead.  The Helper class is only created by the TwoVectorField
// class, so its constructor is private.

class TwoVectorHelper {
private:
  TwoVectorHelper(DegreeOfFreedom *xx, DegreeOfFreedom *yy)
    : x(xx), y(yy)
  {}
  DegreeOfFreedom * const x;
  DegreeOfFreedom * const y;
  friend class TwoVectorFieldBase;
public:
  TwoVectorHelper &operator=(const Coord &v); // field(node) = coord
  operator const Coord() const;	// coord = field(node)
};

class TwoVectorFieldBase : public virtual Field {
public:
  TwoVectorFieldBase(const std::string &name) : Field(name, 2) {}
  virtual ~TwoVectorFieldBase() {}
  virtual DegreeOfFreedom *operator()(const FuncNode*, int component) const;
  virtual DegreeOfFreedom *operator()(const FuncNode &n, int component) const {
    return operator()(&n, component);
  }
  virtual DegreeOfFreedom *operator()(const ElementFuncNodeIterator&,
				      int component) const;
  TwoVectorHelper operator()(const FuncNode*) const;
  TwoVectorHelper operator()(const FuncNode &n) const {
    return operator()(&n);
  }
  TwoVectorHelper operator()(const ElementFuncNodeIterator&) const;
  virtual OutputValue newOutputValue() const;
  virtual OutputValue output(const ElementFuncNodeIterator&) const;
  virtual OutputValue output(const FuncNode&) const;
  virtual void setValueFromOutputValue(const FuncNode&, const OutputValue*);
  virtual IteratorP iterator(Planarity=ALL_INDICES/*irrelevant*/) const;
  virtual IndexP componenttype() const;
  virtual IndexP getIndex(const std::string&) const;
};

class TwoVectorField : public TwoVectorFieldBase, public CompoundField {
private:
  static const std::string classname_;
public:
  TwoVectorField(const std::string &name);
  virtual ~TwoVectorField() {}
  const std::string &classname() const { return classname_; }
};

//------------------------

// See comment above wrt TwoVectorHelper.
class VectorHelper {
private:
  unsigned int ndof;
  DegreeOfFreedom *const*const x;
public:
  VectorHelper(DegreeOfFreedom *const*const xx, int dim) :
    ndof(dim), x(xx)
  {}
  VectorHelper &operator=(const VECTOR_D &); // field(node) = vector
  operator const VECTOR_D() const; // vector = field(node)
};

class VectorFieldBase : public virtual Field {
public:
  VectorFieldBase(const std::string &name, int dim)
    : Field(name, dim)
  {}
  virtual ~VectorFieldBase() {}
  virtual DegreeOfFreedom *operator()(const FuncNode*, int component) const;
  DegreeOfFreedom *operator()(const FuncNode &n, int component) const {
    return operator()(&n, component);
  }
  virtual DegreeOfFreedom *operator()(const ElementFuncNodeIterator&,
				      int component) const;
  VectorHelper operator()(const FuncNode*) const;
  VectorHelper operator()(const FuncNode &n) const {
    return operator()(&n);
  }
  VectorHelper operator()(const ElementFuncNodeIterator&) const;
  virtual OutputValue newOutputValue() const;
  virtual OutputValue output(const ElementFuncNodeIterator&)
    const;
  virtual OutputValue output(const FuncNode&) const;
  virtual void setValueFromOutputValue(const FuncNode&, const OutputValue*);
  virtual IteratorP iterator(Planarity) const;
  virtual IndexP componenttype() const;
  virtual IndexP getIndex(const std::string&) const;
};

Field *getFieldByIndex(int);
int countFields();
CompoundField *getCompoundFieldByIndex(int);
int countCompoundFields();

#endif
