// -*- C++ -*-
// $RCSfile: material.h,v $
// $Revision: 1.53 $
// $Author: langer $
// $Date: 2007/01/25 17:25:30 $

/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

#ifndef MATERIAL_H
#define MATERIAL_H

#include "common/abstractimage.h"
#include "common/ccolor.h"
#include "common/identification.h"
#include "common/pixelattribute.h"
#include "common/timestamp.h"
#include "engine/matvec.h"
#include "engine/sparselink.h"
#include "engine/linearizedsystem.h"
#include <iostream>
#include <string>
#include <vector>

class CColor;
class CSubProblem;
class CMicrostructure;
class Element;
class FEMesh;
class Field;
class Flux;
class FluxData;
class MasterPosition;
class PixelSet;
class Property;
class PropertyOutput;
class StringImage;
class TimeDerivative;

//-\\-//-\\-//-\\-//-\\-//-\\-//-\\-//-\\-//-\\-//-\\-//-\\-//-\\-//

// MaterialPropertyRegistration associates a Property and a tag, such
// as "Elasticity".  A Material can have at most one Property
// associated with any given tag.  The definition is in material.C.

class MaterialPropertyRegistration;

// MaterialPropertyRegistry handles the MaterialPropertyRegistration
// objects for a Material.

class MaterialPropertyRegistry {
private:
  std::vector<MaterialPropertyRegistration*> reg;
public:
  ~MaterialPropertyRegistry();

  // Register a property under the name "tag"
  void registr(Property *prop, const std::string &tag);
  // Find the property registered with the name "tag"
  Property *fetch(const std::string &tag) const;

  void clear();
#ifdef DEBUG
  void dump(std::ostream&) const;
#endif // DEBUG
};

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// class for listing the Properties that contribute to a Flux

class FluxPropList : public std::vector<Property*> {
public:
  // A Flux is active for this material if any of the Properties that
  // contribute to it are active.  A Property is active if all of the
  // Fields it uses are defined.
  bool active(const CSubProblem*) const;
};

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

class Material : public IdentifiedObject {
private:
  std::string name_;
  // consistency_state is empty if the material is consistent.
  std::vector<std::string> consistency_state;


  // all Properties of this Material
  std::vector<Property*> property;
  // list of all Properties that contribute to a Flux
  std::vector<FluxPropList> fluxprop;
  // list of Fields that couple to each Flux via this Material's Properties
  std::vector<std::vector<const Field*> > fluxfield;
  // list of all Properties that contribute to a PropertyOutput
  std::vector<std::vector<Property*> > outputprop;
  void make_fluxfield_table();
  void clear_fluxfield_table();
  void clear_fluxproplist();
  void clear_outputprop();

  TimeStamp timestamp;

  Material(const Material&);	// forbidden
public:
  Material();
  Material(const std::string&);
  ~Material();

  // bool self_consistent() const { return consistent; }
  bool self_consistent() const;
  const std::vector<std::string> *consistency_messages() const;

  const std::string &name() const { return name_; }

  void rename(std::string newname) { name_ = newname; }

  // Properties

  MaterialPropertyRegistry registry; // facilitates interproperty communication
  void registerPropertyType(Property *p, const std::string &name);
  void registerFlux(Property *p, const Flux *flux);
  bool contributes_to_flux(const Flux *flux) const;
  Property *fetchProperty(const std::string &name) const;// uses the property category to retreive the property
  Property *getProperty(int i) const;
  int nProperties() const { return property.size(); }
  
  void add1Property(Property*);
  void remove1Property(Property*);
  void clear_xref();
  void cross_reference();
  void precompute(CSubProblem*); // only precomputes computable properties
  void precompute_all_properties(); // precomputes all Properties, active or not
  void begin_element(CSubProblem*, const Element*) const;
  void end_element(CSubProblem*, const Element*) const;

  // PropertyOutputs
  void registerOutput(Property*, const std::string&);
  const std::vector<Property*> &outputProperties(const PropertyOutput*) const;

  void cleanAttributes(CMicrostructure*) const;
  void replaceAttributes(CMicrostructure*, Material *) const;

  const TimeStamp &getTimeStamp() const;

  // List of fields that contribute to a flux.  If the flux isn't used
  // by this Material, then the vector has no entries.
  const std::vector<const Field*> &fieldlist(const Flux&) const;

  void make_linear_system(CSubProblem*, 
			  const Element*,
			  LinearizedSystem &) const;
  void set_time_derivatives(CSubProblem*, 
			    const Element*,
			    LinearizedSystem&) const;
  void find_fluxmatrix(const FEMesh*, const Element*, const Flux*, FluxData*,
		       const MasterPosition&) const;

  void post_process(CSubProblem*, const Element*) const;

  void assignToPixels(CMicrostructure*, const std::vector<ICoord>*) const;
  void assignToPixelGroup(CMicrostructure*, const PixelSet*) const;
  void assignToAllPixels(CMicrostructure*) const;
  int nPixelsInMicrostructure(const CMicrostructure*) const;
  // Returns a FluxPropList for a give Flux.
  FluxPropList get_flux_props(const Flux*) const;

  friend std::ostream &operator<<(std::ostream &, const Material&);

#ifdef DEBUG
  void dump(std::ostream&) const;
#endif // DEBUG
};


//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//
//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// Classes and functions for handling the assignment of Materials to
// pixels in a Microstructure.

class MaterialAttribute : public PixelAttribute {
private:
  const Material *material;
public:
  MaterialAttribute() : material(0) {}
  void set(const Material *m);
  const Material *get() const { return material; }
  virtual bool operator<(const PixelAttribute&) const;
  virtual bool strictLessThan(const PixelAttribute& x) const {
    return operator<(x);
  }
  virtual void print(std::ostream&) const;
};

class MaterialAttributeRegistration : public PxlAttributeRegistration {
private:
  static const std::string classname_;
  static const std::string modulename_;
public:
  MaterialAttributeRegistration();
  virtual ~MaterialAttributeRegistration() {}
  virtual PixelAttribute *createAttribute(const ICoord&) const {
    // Stored in the calling microstructure's attribute map,
    // destroyed along with the microstructure.
    return new MaterialAttribute();
  }
  virtual PixelAttributeGlobalData*
     createAttributeGlobalData(const CMicrostructure*) const;
  
  virtual const std::string &classname() const { return classname_; }
  virtual const std::string &modulename() const { return modulename_; }
};

const Material *getMaterialFromCategory(const CMicrostructure*, int category);
const Material *getMaterialFromPoint(const CMicrostructure*, const ICoord*);
Array<PixelAttribute*> &getMaterialMap(const CMicrostructure*);
const Array<PixelAttribute*> &getConstMaterialMap(const CMicrostructure*);
std::vector<const Material*> *getMaterials(const CMicrostructure*);
void removeMaterialFromPixels(CMicrostructure*, const PixelSet*);
void removeMaterialFromPixels(CMicrostructure*, const std::vector<ICoord>&);
void removeAllMaterials(CMicrostructure*);
TimeStamp getMaterialTimeStamp(const CMicrostructure*);

class MaterialImage : public AbstractImage {
private:
  CMicrostructure *microstructure;
  const CColor noMaterial;
  const CColor noColor;
public:
  MaterialImage(CMicrostructure*, const CColor*, const CColor*);
  virtual ~MaterialImage() {}
  virtual const Coord &size() const;
  virtual const ICoord &sizeInPixels() const;
  virtual void fillstringimage(StringImage*) const;
};

#endif
