// -*- C++ -*-
// $RCSfile: elasticity.C,v $
// $Revision: 1.45 $
// $Author: langer $
// $Date: 2006/12/19 20:29:47 $

/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

#include <oofconfig.h>
#include "cijkl.h"
#include "common/threadstate.h"
#include "common/coord.h"
#include "common/space.h"
#include "common/trace.h"
#include "elasticity.h"
#include "engine/IO/propertyoutput.h"
#include "engine/cstrain.h"
#include "engine/csubproblem.h"
#include "engine/element.h"
#include "engine/elementnodeiterator.h"
#include "engine/femesh.h"
#include "engine/field.h"
#include "engine/flux.h"
#include "engine/indextypes.h"
#include "engine/material.h"
#include "engine/ooferror.h"

Elasticity::Elasticity(const std::string &nm, PyObject *registration)
  : Property(nm, registration)
{
  displacement = dynamic_cast<TwoVectorField*>(Field::getField("Displacement"));
  stress_flux = dynamic_cast<SymmetricTensorFlux*>(Flux::getFlux("Stress"));
}

void Elasticity::precompute() {
}

int Elasticity::integration_order(const CSubProblem *subp, 
				  const Element *el) const {
  if(displacement->in_plane(subp))
    return el->dshapefun_degree();
  return el->shapefun_degree();
}

void Elasticity::fluxmatrix(const FEMesh *mesh, const Element *element,
			    const ElementFuncNodeIterator &nu,
			    const Flux *flux, FluxData *fluxdata,
			    const MasterPosition &x) const 
{
  // Compute M at point x.
  // stress_{ij} = M_{ij k nu} u_(k nu)  where u_(k nu) is a dof at node nu
  
  // stress_{ij}(x) = C_{ijkl} eps_{kl}(x)
  //                = C_{ijkl} (grad_k u_l + grad_l u_k)/2
  //                = C_{ijkl} grad_k u_l(x)

  // expand u_l(x) =  u_{nu l} N_nu(x)
  //   M_{ij l nu} = C_{ijkl} grad_k N_nu

  if(*flux != *stress_flux) {
    throw ErrProgrammingError("Unexpected flux", __FILE__, __LINE__);
  }

//   Trace("Elasticity::fluxmatrix");
  const Cijkl modulus = cijkl(mesh, element, x);
  double sf = nu.shapefunction(x);
  double dsf0 = nu.dshapefunction(0, x);
  double dsf1 = nu.dshapefunction(1, x);

  // Loop over stress components.  Loop over all components, even if
  // the stress is in-plane, since the out-of-plane components of
  // the flux matrix are used to construct the constraint equation
  // for the out-of-plane stress.
  IteratorP ell = displacement->iterator(); // reuse this;
  for(SymTensorIterator ij; !ij.end(); ++ij) {
    // loop over displacement components for in-plane strain contributions 
    for( ; !ell.end(); ++ell) {
      // loop over k=0,1 is written out explicitly to save a tiny bit of time
      SymTensorIndex ell0(0, ell.integer());
      SymTensorIndex ell1(1, ell.integer());
      fluxdata->matrix_element(ij, displacement, ell, nu) += 
	modulus(ij, ell0)*dsf0 + modulus(ij, ell1)*dsf1;
    }
    ell.reset();
    // loop over out-of-plane strains
    if(!displacement->in_plane(mesh)) {
      Field *oop = displacement->out_of_plane();
      for(IteratorP ell=oop->iterator(ALL_INDICES); !ell.end(); ++ell) {
	double diag_factor = ( ell.integer()==2 ? 1.0 : 0.5);
	fluxdata->matrix_element(ij, oop, ell, nu)
	  += modulus(ij, SymTensorIndex(2, ell.integer())) * sf * diag_factor;
      }
    }
  }
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

void Elasticity::output(const FEMesh *mesh,
			const Element *element,
			const PropertyOutput *output,
			const MasterPosition &pos,
			OutputVal *data)
  const 
{
  const std::string &outputname = output->name();
  if(outputname == "Energy") {
    // The parameter is a Python Enum instance.  Extract its value.
    // The name of the parameter is 'etype', set in outputDefs.py when
    // the ScalarPropertyOutputRegistration for "Energy" was created.
    std::string etype = output->getEnumParam("etype");
    if(etype == "Total" || etype == "Elastic") {
      ScalarOutputVal *edata = 
	dynamic_cast<ScalarOutputVal*>(data);
      SymmMatrix strain(3);
      const Cijkl modulus = cijkl(mesh, element, pos);
      findGeometricStrain(mesh, element, pos, strain);
      SymmMatrix stress(modulus*strain);
      double e = 0;
      for(int i=0; i<3; i++) {
	e += stress(i,i)*strain(i,i);
	int j = (i+1)%3;
	e += 2*stress(i,j)*strain(i,j);
      }
      *edata += 0.5*e;
    }
  }
}

