# -*- python -*-
# $RCSfile: snaprefine.py,v $
# $Revision: 1.8 $
# $Author: langer $
# $Date: 2007/02/16 20:48:54 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

# This file implements "Snap Refine"
# (was called "Intelligent Refine" (pun intended)).
# New edge nodes are transition points.
# The "ruleset" is always liberal.

# Comment from the original refine.py
# Refinement works in two stages.  First a RefinementTarget object (from
# refinementtarget.py) checks the elements and marks edges for
# subdivision.  Then a RefinementRuleSet (from refinemethod.py) is
# used to actually refine the elements, according to how many times
# each edge is marked.

from oof2.SWIG.common.IO import progressbar
from oof2.common import debug
from oof2.common import enum
from oof2.common import parallel_enable
from oof2.common import registeredclass
from oof2.common.IO import parameter
from oof2.common.IO import xmlmenudump
from oof2.engine import refinementtarget
from oof2.engine import refinequadbisection
from oof2.engine import skeletonmodifier
from oof2.engine import skeletonsegment
from oof2.engine import skeletonnode

from oof2.engine import refine
from oof2.engine import snaprefinemethod

SkeletonSegment = skeletonsegment.SkeletonSegment

arbitrary_factor = 128          # used by findSignature().  Must be
				# larger than the largest number of
				# refinements of an edge.

#################################

class EdgeMarkings:
    # Class for storing and returning how many divisions should be
    # performed on an edge.  Edges are defined by a pair of nodes.
    def __init__(self, newSkeleton, oldSkeleton, maxdelta2):
        self.markings = {}
        # New for Snap Refine (was Intelligent Refine)
        self.newEdgeNodes = {}
        self.newSkeleton=newSkeleton
        self.oldSkeleton=oldSkeleton
        self.maxdelta2=maxdelta2
        #RCL: check
       #self.numelements = {}
        
    def mark(self, node0, node1, ndivs):
        # arguments are the nodes defining the edge, and the number of
        # new nodes to add to that edge.
        key = skeletonnode.canonical_order(node0, node1)
        try:
            # Only mark an edge if it's not already marked for more
            # divisions.
            if self.markings[key] < ndivs:
                self.markings[key] = ndivs
        except KeyError:
            self.markings[key] = ndivs

        # mark the partner segment, if it exists
        partners = node0.getPartnerPair(node1)
        if partners is not None:
            partnerKey = skeletonnode.canonical_order(partners[0], partners[1])
            try:
                # Only mark an edge if it's not already marked for more
                # divisions.
                if self.markings[partnerKey] < ndivs:
                    self.markings[partnerKey] = ndivs
            except KeyError:
                self.markings[partnerKey] = ndivs

    def getMark(self, node0, node1):
        key = skeletonnode.canonical_order(node0, node1)
        try:
            return self.markings[key]
        except KeyError:
            return 0

    def getMarks(self, element):
        nnodes = element.nnodes()
        return [self.getMark(element.nodes[i], element.nodes[(i+1)%nnodes])
                for i in range(nnodes)]

    #RCL: check
    #def InitNumElements(self):
    #    for seg in self.markings.keys():
    #        self.numelements[seg]=0
        
    # Modified from refine.py to take into account the
    # edge categories and transition points.
    # This method does the work of both getMarks and getNewEdgeNodes
    # This method selects which marked edges are to be bisected.
    def getSnapMarks(self, element):
        newSkeleton=self.newSkeleton
        numinitmarks=0
        intelmarks=[]
        intelcats=[]
        inteltranspts=[]
        nnodes = element.nnodes()
        for i in range(nnodes):
            node0=element.nodes[i]
            node1=element.nodes[(i+1)%nnodes]
            n0pt=node0.position()
            n1pt=node1.position()
            key=skeletonnode.canonical_order(node0,node1)
            initmark=0
            notmarked=True
            try:
                initmark=self.markings[key]
                numinitmarks+=1
            except KeyError:
                initmark=-1
            if initmark>0:
                transpt=None
                transptnode=None
                try:
                    transptnode = self.newEdgeNodes[key]
                    if transptnode:
                        transpt=transptnode.position()
                except KeyError:
                    # Get a transition point starting from each end
                    transpt0 = newSkeleton.MS.transitionPointWithPoints(n0pt,n1pt)
                    transpt1 = newSkeleton.MS.transitionPointWithPoints(n1pt,n0pt)
                    # If a transition point exists from one end,
                    # it should exist from the other (may not be true in case
                    # the edge is at the interface of pixel groups)
                    # Also need to check if a transition point coincides with
                    # either endpoint (it happens!)
                    gettranspt0=False
                    if transpt0:
                        d0=(transpt0-n0pt)**2
                        d1=(transpt0-n1pt)**2
                        gettranspt0=(d0>self.maxdelta2 and d1>self.maxdelta2)
                    gettranspt1=False
                    if transpt1:
                        d1=(transpt1-n1pt)**2
                        d0=(transpt1-n0pt)**2
                        gettranspt1=(d1>self.maxdelta2 and d0>self.maxdelta2)
                    if gettranspt0 and gettranspt1:
                        # Pick the transition point closest to the midpoint
                        midpt=0.5*(n0pt+n1pt)
                        if (transpt0-midpt)**2 < (transpt1-midpt)**2:
                            transpt=transpt0
                        else:
                            transpt=transpt1
                    elif gettranspt0:
                        transpt=transpt0
                    elif gettranspt1:
                        transpt=transpt1
                if transpt:
                    homog0, cat0 = newSkeleton.MS.edgeHomogeneityCat(transpt,n0pt)
                    homog1, cat1 = newSkeleton.MS.edgeHomogeneityCat(transpt,n1pt)
                    if 1:#cat0 != cat1:
                        intelmarks.append(1)
                        intelcats.append((cat0,cat1))
                        notmarked=False
                        if not transptnode:
                            transptnode = newSkeleton.newNode(transpt.x,transpt.y)
                            # Instead of a list of new nodes (see refine.py),
                            # a transition point is assigned as a new edge node
                            self.newEdgeNodes[key] = transptnode
                        #self.numelements[key]=self.numelements[key]+1
                        inteltranspts.append([transptnode])
            if notmarked:
                if initmark>-1: #initially marked but was not bisected
                    self.newEdgeNodes[key]=None
                    #self.numelements[key]=self.numelements[key]+1
                intelmarks.append(0)
                homogedge,catedge=newSkeleton.MS.edgeHomogeneityCat(n0pt,n1pt)
                #if homogedge<0.5:
                #    catedge=-1
                intelcats.append((catedge,))
                inteltranspts.append([])

            try:
                transptnode = self.newEdgeNodes[key]
                partners = node0.getPartnerPair(node1)
                if partners is not None:
                    partnerKey = skeletonnode.canonical_order(partners[0], partners[1])
                    # Set the partners of the new edge nodes.  Partners of the
                    # old nodes were set by improperCopy.  Given the nodes at
                    # the ends of the segment being divided in the
                    # oldSkeleton, get the partner pair, the children of the
                    # partner pair, the segment between the partner pair, and
                    # the children of that segment, if they have been created
                    # yet. For each child segment, get the endpoints and
                    # partner them with the new nodes on the segment we are
                    # currently working on.

                    if transptnode:
                        # Map the transition point to the correct coordinate
                        # at the opposite boundary
                        sx=newSkeleton.MS.size()[0]
                        sy=newSkeleton.MS.size()[1]
                        tx=transptnode.position()[0]
                        ty=transptnode.position()[1]
                        # We expect transptnode to be at a (periodic) boundary
                        # because the node0 and node1 have partners
                        if tx==0:
                            tx=sx
                        elif tx==sx:
                            tx=0
                        elif ty==0:
                            ty=sy
                        elif ty==sy:
                            ty=0
                        transptnodepartner=newSkeleton.newNode(tx,ty)
                        self.newEdgeNodes[partnerKey]=transptnodepartner
                        #transptnode.addPartner(transptnodepartner)
                        # child of partner of node0
                        childNode0 = partners[0].getChildren()[-1]
                        # child of partner of node1
                        childNode1 = partners[1].getChildren()[-1] 
                        parentSegment = self.oldSkeleton.findSegment(partners[0],partners[1])
                        childrenSegments=None
                        if parentSegment:
                            childrenSegments = parentSegment.getChildren()
                        else:
                            import pdb
                            pdb.set_trace()
                        # TODO: Eliminate rounding error in positions of new
                        # partners
                        for child in childrenSegments:
                            csnode0, csnode1 = child.nodes()
                            if csnode0 == childNode0 or csnode1 == childNode0:
                                child.get_other_node(childNode0).addPartner(transptnode)
                            # in case it is a trisection, partner up the
                            # second added node, if it's a bisection, this
                            # does nothing
                            #if csnode0 == childNode1 or csnode1 == childNode1:
                            #    child.get_other_node(childNode1).addPartner(nodes[-1])
                    else:
                        self.newEdgeNodes[partnerKey]=None
            except KeyError:
                pass

        return (numinitmarks, intelmarks, intelcats, inteltranspts)

    def getNMarkedEdges(self, element):
        marks = self.getMarks(element)
        nmarks = 0
        for m in marks:
            if m: nmarks += 1
        return nmarks

##class RefinementCriterion(registeredclass.RegisteredClass):
##    registry = []

##    tip = "Restrict the set of Elements considered for refinement."
##    discussion = """<para>

##    Objects in the <classname>RefinementCriterion</classname> class
##    are used as values of the <varname>criterion</varname> parameter
##    when <link linkend='RegisteredClass:Refine'>refining</link>
##    &skels;.  Only &elems; that satisfy the criterion are considered
##    for refinement.

##    </para>"""


##class Unconditionally(RefinementCriterion):
##    def __call__(self, skeleton, element):
##        return 1
##registeredclass.Registration(
##    'Unconditional',
##    RefinementCriterion,
##    Unconditionally,
##    ordering=0,
##    tip='Consider all Elements for possible refinement.',
##    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/reg/unconditionally.xml')
##    )

##class AreaUnits(enum.EnumClass(
##    ('Pixel', 'Area is in units of the pixel area.'),
##    ('Physical', 'Area is in physical units.'))):
##    tip = "Units of area."
##    discussion = """<para>
##    Units of element area.  See <xref
##    linkend='RegisteredClass:MinimumArea'/> for details.
##    </para>"""

##class MinimumArea(RefinementCriterion):
##    def __init__(self, threshold, units):
##        self.threshold = threshold
##        self.units = units
##    def __call__(self, skeleton, element):
##        if self.units == 'Pixel':
##            return element.area() > self.threshold*skeleton.MS.areaOfPixels()
##        else:
##            return element.area() > self.threshold

##registeredclass.Registration(
##    'Minimum Area',
##    RefinementCriterion,
##    MinimumArea,
##    ordering=1,
##    params=[parameter.FloatParameter('threshold', 10,
##                                     tip="Minimum acceptable element area."),
##            enum.EnumParameter('units', AreaUnits, AreaUnits('Pixel'),
##                               tip='Units for the minimum area')],
##    tip='Only refine elements with area greater than the given threshold.',
##    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/reg/minimumarea.xml')
##    )

###################################
            
class SnapRefine(skeletonmodifier.SkeletonModifier):
    #Unlike refine.py, alpha parameter not used (yet?)
    def __init__(self, targets, criterion, min_distance, alpha=1):
        self.targets = targets      # RefinementTarget instance
        self.criterion = criterion  # Criterion for refinement
        # alpha is used for deciding between different possible
        # refinements, using effective energy
        self.alpha = alpha 
        # Available rules
        self.rules = snaprefinemethod.getRuleSet('liberal')

        self.min_distance=min_distance
    #########

    def applyAMR(self, skeleton, context): # for adaptive mesh refinement
        newSkeleton = skeleton.improperCopy(fresh=True)
        return self.refinement(skeleton, newSkeleton, context)
    
    def apply(self, skeleton, context):
        newSkeleton = skeleton.improperCopy()  # fresh=False
        return self.refinement(skeleton, newSkeleton, context)

    def refinement(self, skeleton, newSkeleton, context):
        # pBar.set_message("marking elements...       ")
        pBar = progressbar.getProgress()

        (maxdelta,deltay)=newSkeleton.MS.sizeOfPixels()
        if maxdelta<deltay:
            maxdelta=deltay
        maxdelta2=(self.min_distance*maxdelta)**2
        markedEdges = EdgeMarkings(newSkeleton,skeleton,maxdelta2)
        self.newEdgeNodes = {}          # allows sharing of new edge nodes

        # Primary marking (bisections only!)
        self.targets(skeleton, context, 1, markedEdges,
                     self.criterion)

        #RCL: check
        #markedEdges.InitNumElements()

        # Additional marking
        #self.degree.markExtras(skeleton, markedEdges)

        # Refine elements and segments
        segmentdict = {}                # which segments have been handled
        elements = skeleton.elements
        n = len(elements)
        
        for ii in range(n):
            oldElement = elements[ii]
            oldnnodes = oldElement.nnodes()
            # Get list of number of subdivisions on each edge ("marks")
            (numinitmarks,marks,cats,edgenodes) = \
                          markedEdges.getSnapMarks(oldElement)

            # Find the canonical order for the marks. (The order is
            # ambiguous due to the arbitrary choice of the starting
            # edge.  Finding the canonical order allows the refinement
            # rule to be found in the rule table.)  rotation is the
            # offset into the elements node list required to match the
            # refinement rule to the element's marked edges.
            # signature is the canonical ordering of the marks.
            rotation, signature = findSignature(marks)
            
            # Create new nodes along the subdivided element edges
##            edgenodes = [self.getNewEdgeNodes(oldElement.nodes[i],
##                                              oldElement.nodes[(i+1)%oldnnodes],
##                                              marks[i], newSkeleton)
##                         for i in range(oldnnodes)]

            # Create new elements
            if numinitmarks>0:
                newElements = self.rules[signature].apply(oldElement, rotation,
                                                          cats,
                                                          edgenodes, newSkeleton,
                                                          maxdelta2)
            else:
                newElements = snaprefinemethod.unrefinedelement(oldElement,
                                                                 rotation,
                                                                 newSkeleton)
            # If the old element's homogeneity is 1, it's safe to say that
            # new elements' homogeneities are 1.
            if oldElement.homogeneity(skeleton.MS) == 1.0:
                for el in newElements:
                    el.copyHomogeneity(oldElement)

            # The calls to Skeleton.newElement() made by the
            # refinement rules have created new SkeletonSegments in
            # newSkeleton, but have not set the parentage of those
            # segments.  We have to fix that here.
            for newElement in newElements:
                for segment in newElement.getSegments(newSkeleton):
                    #RCL: check
                    #if segment.length()==0.0:
                    #    import pdb
                    #    pdb.set_trace()
                    # Only look at each segment once.
                    if not segmentdict.has_key(segment):
                        segmentdict[segment] = 1
                        pseg = findParentSegment(skeleton, newElement,
                                                 segment,
                                                 edgenodes)
                        if pseg:
                            pseg.add_child(segment)
                            segment.add_parent(pseg)
            if pBar.query_stop():
                pBar.set_failure()
                pBar.set_message("Failed")
                return None
            else:
                pBar.set_progress(1.0*(ii+1)/n)
                pBar.set_message("refining skeleton: %d/%d" % (ii+1, n))

        
        newSkeleton.cleanUp()
        #RCL: check
##        for e,n in markedEdges.numelements.items():
##            if n!=0 and n!=2:
##                x0=e[0].position().x
##                x1=e[1].position().x
##                y0=e[0].position().y
##                y1=e[1].position().y
##                if (x0==0 and x1==0) or (y0==0 and y1==0)\
##                      or (x0==353 and x1==353) or (y0==221 and y1==221):
##                    pass
##                else:
##                    print e,n, e[0].position(), e[1].position()
        return newSkeleton

    ################

##    def getNewEdgeNodes(self, node0, node1, ndivs, newSkeleton):
##        # Create ndivs new nodes between node0 and node1 in the
##        # skeleton.
##        if ndivs < 1:
##            return []
##        key = skeletonnode.canonical_order(node0, node1)
##        try:
##            nodes = self.newEdgeNodes[key]
##            # Since this is the second time we're using this list of
##            # nodes, we must be looking at them from the other side,
##            # and the nodes should be in the opposite order.
##            # Reversing them in place like this would be wrong if the
##            # list weren't being used immediately, as it is in
##            # Refine.apply()
##            nodes.reverse()
##        except KeyError:
##            nodes = [None]*ndivs
##            p0 = node0.position()
##            p1 = node1.position()
##            delta = (p1 - p0)/(ndivs+1)
##            for i in range(ndivs):
##                pt = p0 + (i+1)*delta
##                newnode = newSkeleton.newNode(pt.x, pt.y)
##                nodes[i] = newnode
##            self.newEdgeNodes[key] = nodes
##        return nodes

#################################

def findParentSegment(oldSkeleton, element, segment, edgenodes):
    n0, n1 = segment.nodes()            # nodes of the segment
    parents = n0.getParents()
    if parents:
        p0 = parents[-1]
    else:
        p0 = None
        
    parents = n1.getParents()
    if parents:
        p1 = parents[-1]
    else:
        p1 = None    
    oldElement = element.getParents()[-1]
    nnodes = oldElement.nnodes()

    if p0 is not None and p1 is not None:
        pseg = oldSkeleton.findSegment(p0, p1)
        if pseg:
            # Trivial case: the parent nodes define a segment in the
            # parent skeleton.  That segment is this segment's parent.
            return pseg
        # Both nodes have parents, but the parent nodes don't define a
        # segment in the old skeleton.  (This is unlikely to happen
        # during refinement.)  The segment does not have a parent.
        return None

    elif p0 is not None or p1 is not None:
        # Exactly one new endpoint has a parent in the old element.
        # The other endpoint may lie on an old segment.  If that
        # segment's other node is the parent of the new node, then
        # that segment is the parent of this segment.
        if p0 is None:
            parentnode = p1
            childnode = n1
            freechild = n0
        else:
            parentnode = p0
            childnode = n0
            freechild = n1
        which = oldElement.nodes.index(parentnode)
        if freechild in edgenodes[which]:
            next = (which+1)%nnodes
            return oldSkeleton.findSegment(parentnode, oldElement.nodes[next])
        prev = (which - 1 + nnodes) % nnodes
        if freechild in edgenodes[prev]:
            return oldSkeleton.findSegment(parentnode, oldElement.nodes[prev])

    elif p0 is None and p1 is None:
        # Neither new endpoint has a parent.  If the new endpoints are
        # consecutive nodes in the list of new nodes added to a parent
        # element segment, then that parent segment is the parent
        # we're looking for.
        for i in range(nnodes):
            nodelist = edgenodes[i]
            try:
                which0 = nodelist.index(n0)
                which1 = nodelist.index(n1)
            except ValueError:
                pass
            else:
                if abs(which0 - which1) == 1:
                    return oldSkeleton.findSegment(oldElement.nodes[i],
                                                   oldElement.nodes[(i+1)%
                                                                    nnodes])

################################

def findSignature(marks):
    # Given a list of subdivisions of sides (marks), rotate a
    # canonical starting point for the list, so that it can be used as
    # an index into a table of refinement functions.  For a list
    # [x,y,z], the canonical order is that which maximizes the number
    # xyz in base arbitrary_factor.  This will fail if edges are ever
    # divided into more than arbitrary_factor segments in one
    # refinement operation.
    n = len(marks)
    max = -1
    imax = None
    for i in range(n):
        key = marks[i]
        for j in range(1, n):
            key = arbitrary_factor*key + marks[(i+j)%n]
        if key > max:
            max = key
            imax = i
    return imax, tuple([marks[(i+imax)%n] for i in range(n)])


####################

registeredclass.Registration(
    'Snap Refine',
    skeletonmodifier.SkeletonModifier,
    SnapRefine,
    ordering=400,
    params=[parameter.RegisteredParameter('targets',
                                          refinementtarget.RefinementTarget,
                                          tip='Target elements to be refined.'),
            parameter.RegisteredParameter('criterion', refine.RefinementCriterion,
                                          tip='Exclude certain elements.'),
            parameter.FloatParameter('min_distance', value=1.0,
                                     tip='Minimum distance of the transition point from either end-point, in units of the maximum pixel dimension.')
#            ,skeletonmodifier.alphaParameter
            ],
    tip="Subdivide elements along pixel boundaries.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/reg/snaprefine.xml')
    )
